#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#if !defined(powerpc)
  #include <math.h>
#else
  #include <fp.h>
  #include <fenv.h>
#endif
#if defined(UNIX) || defined(WIN32)
  #include <sys/stat.h>
  #include <time.h>
#else
  #include <stat.h>
#endif
#ifndef size_t
  #define size_t unsigned
#endif

#include "allocate.h"
#include "color.h"
#include "imageio.h"
#include "heap.h"

#define nbfeature		9
#define nbweight		9

/*
#define nbfeature	( HIST0_BINS + HIST1_BINS + HIST2_BINS + HIST3_BINS + HIST4_BINS + HIST5_BINS + HIST6_BINS + HIST7_BINS + HIST8_BINS)
#define nbweight	( HIST0_BINS + HIST1_BINS + HIST2_BINS + HIST3_BINS + HIST4_BINS + HIST5_BINS + HIST6_BINS + HIST7_BINS + HIST8_BINS)
*/

#define MAXVALUE		((float) 1e+30)
#define MAXLEVEL		256

#define COLORSPACE   		LAB
#define GAMMA				2.2

#define HIST0_BINS              8
#define HIST1_BINS              16
#define HIST2_BINS              16
#define HIST3_BINS              16 
#define HIST4_BINS              8
#define HIST5_BINS              8
#define HIST6_BINS              9
#define HIST7_BINS              9
#define HIST8_BINS              9

struct feature
{
  int   pts;
  int   status;
  int   row, col;
  float hist0[HIST0_BINS];
  float hist1[HIST1_BINS];
  float hist2[HIST2_BINS];
  float hist3[HIST3_BINS];
  float hist4[HIST4_BINS];
  float hist5[HIST5_BINS];
  float hist6[HIST6_BINS];
  float hist7[HIST7_BINS];
  float hist8[HIST8_BINS];
};

struct float_img
{
  int   col,row;
  float **mono;
  float ***color;
};

struct node
{
  int k;            /* image index of node */
  int key;
};               

struct node_set {
  struct heap h;      /* heap used to find best location node */
  struct node *p;     /* pointer to array of location nodes */
  int N_nodes;        /* number of nodes that have been used */
  int alloc;          /* number of nodes that have been allocated */
};

struct img_info
{
  char name[80];
  int  type;
  int  ix;
  int  num_img;
  int  col,row;      		/* number of columns and rows in base image */
  int  depth, level, x, y;
  int  order, rank;
  int  searchcount;
  long featurept;
  long featurelen;
  long thumbpt;
  double dist, size;
  struct feature *feature;
  struct img_info *next, *child, *key, *nearest;
};

struct img_set {
  int num_img;
  int max_img;
  struct feature *featureblock;
  struct img_info *img, *root;
};

struct info_level
{
  int    level, col, row;
  struct img_info ***info;
};

struct info_pym {
  int    num_level;
  struct info_level level[MAXLEVEL];
};

/* io.c */
void   init_img_info  (struct img_info *img_info);
void   copy_img_info  (struct img_info *img_info, struct img_info *new_info);
void   allocate_img_info(struct img_set *key_set, int num_img);
void   free_img_info  (struct img_set *img_set);
void   get_img_info   (struct img_set *img_set, char *dbname,
                       char *dirname, int num_img, int db_type);
void   write_img_info (struct img_set *img_set, char *dbname, char *dirname);
void   append_img_info(struct img_set *img_set, struct img_info *img_info);

void   allocate_img_set(struct img_set *img_set);
void   read_img_set   (struct img_set *img_set, char *dbname,
                       char *subdir,  char *mode);
void   write_img_set  (struct img_set *img_set, char *dbname,
                       char *subdir,  char *mode);

void   get_img_tree   (struct img_set *img_set, struct img_set *key_set, 
                       char *dbname, char *subdir);
void   write_img_tree (struct img_set *img_set, struct img_set *key_set, 
                       char *dbname, char *subdir);

void   read_info_pym  (struct img_set  *img_set,  struct img_set *key_set,
                       struct info_pym *info_pym, char *dbname, char *subdir);
void   write_info_pym (struct img_set  *img_set,  struct img_set *key_set,
                       struct info_pym *info_pym, char *dbname, char *subdir);

/* util.c */
void write_fimg(char *fname, struct float_img *img1);
void copy_img (struct img *img1, struct img *img2);
int  thumbnail(struct img *data_img, struct img *thumb_img, int maxsize);
void prepare_catalog(struct img *all_img, int row, int col, float ratio);
void copy_to_catalog(struct img *img1, struct img *img2, int row, int col, float ratio);
void copy_to_catalog2(struct img *img1, struct img *img2, int row, int col, float ratio, int boxcolor);
int  mono_to_color(struct img *data_img, int img_type);
void color_space_transform(struct img *img, struct float_img *fimg);
double image_distance  (struct float_img *dimg, struct float_img *timg);

/* feature.c */
double feature_distance(struct feature *dfeature, struct feature *tfeature); 
void   build_feature(struct float_img *fimg, struct feature *feature); 
void   naturalize(struct feature *feature);
void   setFilter(float sigma);
void   setWeight(float *newweight);
int    write_feature (struct feature *feature, FILE *fp, int *lstart, int *len);
int    read_feature  (struct feature *feature, FILE *fp, int  lstart, int  len);
int    read_feature2 (struct feature *feature, FILE *fp, int  lstart, int  len);
void   free_feature  (struct feature *feature);
void   free_feature2 (struct feature *feature);
int    dump_feature  (struct feature *feature);
int    reset_feature (struct feature *feature);
int    copy_feature  (struct feature *feature, struct feature *newfeature);
int    add_feature   (struct feature *feature, struct feature *newfeature, float ratio);
int    adjust_feature(struct feature *feature, struct feature *newfeature, float ratio);
struct feature *new_feature  (void);
FILE   *open_database (char *dbname, char *subdir, char *mode);
void   close_database (FILE *fp);
void   setWeight      (float *weight);
void   setFilter      (float sigma);


/* search.c */
void fast_db_search(
  struct img_set  *img_set,
  struct img_set  *key_set,
  struct node_set *best_ns,
  struct img_info *timg_info,
  int             num_to_return);

void list_db_search(
  struct img_set  *img_set,
  struct img_set  *key_set,
  struct img_info *root,
  struct node_set *best_ns,
  struct img_info *timg_info,
  int             num_to_return);

void list_linear_db_search(
  struct img_set  *img_set,
  struct img_set  *key_set,
  struct img_info *root,
  struct node_set *best_ns,
  struct img_info *timg_info,
  struct img_info **infoqueue);

double direct_search(
  struct img_info *dimg_info, 
  struct img_info *timg_info, 
  struct node_set *ns);

void setHeurist(float newrsize);


/* node_util.c */
void initialize_node_set   (struct node_set *ns);
void free_node_set         (struct node_set *ns);
void reset_node_set        (struct node_set *ns);
void insert_into_node_set  (struct node_set *ns, struct node *p, double dist);
int  get_best_from_node_set(struct node_set *ns, struct node *p, double *dist);

/* parse.c */
char *basename(char *fullname);
void cmd_parse(int argc, char *argv[], 
               char *tpl_name, char *path_name, char *dir_name,
	       int *minimax, int *num_to_find, int *test, int *subset,
	       float *weight, float *filter, float *rsize);

/* routine for VQ tree partition */
struct img_info *partition(struct img_set  *img_set, struct img_set *key_set,  
                           struct img_info *img_list, int level);
int   prunebylist(struct img_set *img_set, struct img_set *sub_set, 
				  int *submap, int nbtpl, int *tpl_ix, int kth);
int   relevance(struct img_set *img_set, int nbtpl, int *tpl_ix, float *weight); 
int   buildtree1(struct img_set *img_set, struct img_set *search_set, struct img_set *binary_set);
int   buildtree2(struct img_set *img_set, struct img_set *binary_set);
int   buildtree3(struct img_set *img_set, struct img_set *key_set, int nbrel, int *relevance);
void *regroup(struct img_set  *img_set, struct img_set  *key_set,
			  struct img_set  *new_set, struct img_info *root,
			    int    level, int nbrel, int *relevance);
int   mapping(struct img_set *img_set);

