#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>
#include <math.h>

#include "imageio.h"
#include "jpeg.h"
#include "pxm.h"
#include "tiff.h"


/*
void main (int argc, char **argv)
{
  int type;
  struct img testimg;

  if(argc!=3) {
    fprintf(stderr,"Wrong number of arguements for program %s\n",argv[0]);
    exit (-1);
  }
  read_img(argv[1], &testimg, &type);
  write_img(argv[2], &testimg, PPM_IMAGE);

}
*/

int write_img(char *fname, struct img *img, int option)
{
  FILE *fp;
  struct pxm_img pxm_img;
  struct TIFF_img tiff_img;
  struct JPEG_img jpeg_img;


  if((fp=fopen(fname,"wb"))==NULL) {
    fprintf(stderr,"\nError: can't write to %s\n",fname); 
    return(FAILED);
  }

  if( option==PBM_IMAGE ) {

    pxm_img.width = img->col;
    pxm_img.height = img->row;
    pxm_img.mono = img->mono;
    pxm_img.max_gray = 1;
    pxm_img.pxm_type = 'b';

    if( write_pxm(fp,&pxm_img)==1 ) {
      fprintf(stderr,"\nError: can't write to PBM image %s\n",fname); 
      fclose(fp);
      return(FAILED);
    }
    fclose(fp);
  }

  if( option==PGM_IMAGE ) {
    pxm_img.width = img->col;
    pxm_img.height = img->row;
    pxm_img.mono = img->mono;
    pxm_img.max_gray = 255;
    pxm_img.pxm_type = 'g';

    if( write_pxm(fp,&pxm_img)==1 ) {
      fprintf(stderr,"\nError: can't write to PGM image %s\n",fname); 
      fclose(fp);
      return(FAILED);
    }
    fclose(fp);
  }

  if( option==PPM_IMAGE ) {
    pxm_img.width = img->col;
    pxm_img.height = img->row;
    pxm_img.color = img->color;
    pxm_img.pxm_type = 'p';

    if( write_pxm(fp,&pxm_img)==1 ) {
      fprintf(stderr,"\nError: can't write to PPM image %s\n",fname); 
      fclose(fp);
      return(FAILED);
    }
    fclose(fp);
  }

  if( option==GRAY_TIFF_IMAGE ) {
    tiff_img.width = img->col;
    tiff_img.height = img->row;
    tiff_img.TIFF_type = 'g';
    tiff_img.compress_type = 'u';
    tiff_img.mono = img->mono;

    if( write_TIFF(fp,&tiff_img)==1 ) {
      fprintf(stderr,"\nError: can't write to TIFF image %s\n",fname); 
      fclose(fp);
      return(FAILED);
    }
    fclose(fp);
  }

  if( option==COLOR_TIFF_IMAGE ) {
    tiff_img.width = img->col;
    tiff_img.height = img->row;
    tiff_img.TIFF_type = 'c';
    tiff_img.compress_type = 'u';
    tiff_img.color = img->color;

    if( write_TIFF(fp,&tiff_img)==1 ) {
      fprintf(stderr,"\nError: can't write to TIFF image %s\n", fname); 
      fclose(fp);
      return(FAILED);
    }
    fclose(fp);
  }

  if( option==GRAY_JPEG_IMAGE ) {
    jpeg_img.width = img->col;
    jpeg_img.height = img->row;
    jpeg_img.jpeg_type = 'g';
    jpeg_img.mono = img->mono;

    if( write_JPEG(fp,&jpeg_img, 75, 0)==1 ) {
      fprintf(stderr,"\nError: can't write to JPEG image %s\n", fname); 
      fclose(fp);
      return(FAILED);
    }
    fclose(fp);
  }

  if( option==COLOR_JPEG_IMAGE ) {
    jpeg_img.width = img->col;
    jpeg_img.height = img->row;
    jpeg_img.jpeg_type = 'c';
    jpeg_img.color = img->color;

    if( write_JPEG(fp,&jpeg_img, 95, 0)==1 ) {
      fprintf(stderr,"\nError: can't write to JPEG image %s\n", fname); 
      fclose(fp);
      return(FAILED);
    }
    fclose(fp);
  }

  return(SUCCESSED);
}




int read_img(char *fname, struct img *img, int *type)
{
  FILE *fp;
  struct pxm_img pxm_img;
  struct TIFF_img tiff_img;
  struct JPEG_img jpeg_img;
  int status;

  status = 0;

  if((fp=fopen(fname,"rb"))==NULL) {
    fprintf(stdout,"Error: can't read from %s\n",fname); 
    return(FAILED);
  }

  if( read_pxm(fp, &pxm_img)==0 ) {
    status = 1;

    if( pxm_img.pxm_type == 'p') {
      *type = PPM_IMAGE;
      img->col = pxm_img.width;
      img->row = pxm_img.height;
      img->color = pxm_img.color;
    }

    if( pxm_img.pxm_type == 'g') {
      *type = PGM_IMAGE;
      img->col = pxm_img.width;
      img->row = pxm_img.height;
      img->mono = pxm_img.mono;
    }

    if( pxm_img.pxm_type == 'b') {
      *type = PBM_IMAGE;
      img->col = pxm_img.width;
      img->row = pxm_img.height;
      img->mono = pxm_img.mono;
    }

  } else if (read_TIFF(fp, &tiff_img)==0) {

    if( tiff_img.TIFF_type == 'g') {
      *type = GRAY_TIFF_IMAGE;
      img->col = tiff_img.width;
      img->row = tiff_img.height;
      img->mono = tiff_img.mono;
    }

    if( (tiff_img.TIFF_type == 'c') ) {
      *type = COLOR_TIFF_IMAGE;
      img->col = tiff_img.width;
      img->row = tiff_img.height;
      img->color = tiff_img.color;
    }

  } else if (read_JPEG(fp, &jpeg_img, 0)==0) {

    if( jpeg_img.jpeg_type == 'g') {
      *type = GRAY_JPEG_IMAGE;
      img->col = jpeg_img.width;
      img->row = jpeg_img.height;
      img->mono = jpeg_img.mono;
    }

    if( (jpeg_img.jpeg_type == 'c') ) {
      *type = COLOR_JPEG_IMAGE;
      img->col = jpeg_img.width;
      img->row = jpeg_img.height;
      img->color = jpeg_img.color;
    }

  } else {
    fprintf(stdout,"Error: can not recognize format of %s\n",fname); 
    fclose(fp);
    return(FAILED);

  }

  fclose(fp);
  return(SUCCESSED);
}

