#include "defs.h"
#undef  LATEX
#define DEFSIZE		1000

void init_img_info(struct img_info *img_info)
{

  strcpy(img_info->name, "NEW");
  img_info->ix   = 0;
  img_info->type = 0;
  img_info->row  = 0;
  img_info->col  = 0;
  img_info->num_img = 0;
  img_info->next = NULL;
  img_info->child= NULL;
  img_info->key  = NULL;
  img_info->nearest = NULL;
  img_info->featurept = 0;
  img_info->featurelen = -1;
  img_info->thumbpt = 0;
  img_info->searchcount = 0;
 
}

void copy_img_info(struct img_info *img_info, struct img_info *new_info)
{

  strcpy(img_info->name, new_info->name);

  img_info->type = new_info->type;
  img_info->row  = new_info->row;
  img_info->col  = new_info->col;
  img_info->num_img = new_info->num_img;
  img_info->searchcount = 0;

}

void allocate_img_info(struct img_set *img_set, int num_img)
{

  img_set->num_img = 0;
  if (num_img < DEFSIZE) img_set->max_img = DEFSIZE;
  else					 img_set->max_img = num_img;
  img_set->img = (struct img_info *) mget_spc(img_set->max_img, sizeof(struct img_info));
  img_set->featureblock = NULL;

}

void free_img_info(struct img_set *img_set)
{

  if (img_set->featureblock != NULL) free( (void *) img_set->featureblock);
  img_set->featureblock = NULL;

  if (img_set->img != NULL) free( (void *) img_set->img);
  img_set->img = NULL;

}

void get_img_info(struct img_set *img_set, char *dbname,
                  char *dirname, int num_img, int db_type)
{
  FILE *fp;
  int k, row, col;
  char buffer[512], longname[256];


  img_set->num_img = 0;
  img_set->max_img = 0;
  img_set->img = NULL;
  img_set->featureblock = NULL;

#if defined(UNIX) || defined(WIN32)
  sprintf(longname, "%s/%s/%s",  dbname, dirname, "db_index");
#else
  sprintf(longname, ":%s:%s:%s", dbname, dirname, "db_index");
#endif

  if ((fp=fopen(longname, "r"))==NULL) {
	if (num_img < DEFSIZE) img_set->max_img = DEFSIZE;
	else				   img_set->max_img = num_img;
    img_set->img = (struct img_info *) mget_spc(img_set->max_img, sizeof(struct img_info));
    return;
  }

  if(fscanf(fp,"%d\n", &(img_set->num_img))!=1) {
    fprintf(stderr,"\nError: Corrupted file %s\n", longname); 
    exit(1);
  }

  img_set->max_img = img_set->num_img + num_img;

  img_set->img = (struct img_info *) mget_spc(img_set->max_img, sizeof(struct img_info));

  /* get the name of each image in the database */
  for(k=0; k<img_set->num_img; k++) {

    /* get image name */
    if (fgets(buffer, 80, fp) == NULL) {
      fprintf(stderr,"\nError1: Corrupted file %s\n", longname);
      exit(1);
    }
    strcpy(img_set->img[k].name, buffer);
    img_set->img[k].name[strlen(buffer)-1] = '\0';

    if(fscanf(fp,"%d %d", &row, &col) != 2) {
      fprintf(stderr,"\nError2: Corrupted file %s\n", longname);
      exit(1);
    }

    img_set->img[k].ix   = k;
    img_set->img[k].type = db_type;
    img_set->img[k].row  = row;
    img_set->img[k].col  = col;
    img_set->img[k].num_img = 0;
    img_set->img[k].level   = -1;
    img_set->img[k].x       = -1;
    img_set->img[k].y       = -1;
    img_set->img[k].feature  = NULL;
    img_set->img[k].next = NULL;
    img_set->img[k].child= NULL;
    img_set->img[k].key  = NULL;
    img_set->img[k].nearest = NULL;
    img_set->img[k].size = 0;
    img_set->img[k].dist = 0;

    if (fscanf(fp," %ld %ld %ld\n", &(img_set->img[k].featurept),
                                    &(img_set->img[k].featurelen),
                                    &(img_set->img[k].thumbpt)) != 3) {
      fprintf(stderr,"\nError3: Corrupted file %s\n", longname);
      exit(1);
    }
    img_set->img[k].searchcount = 0;
  }

  for(k=img_set->num_img; k<img_set->max_img; k++) {
    img_set->img[k].ix   = k;
    img_set->img[k].feature  = NULL;
  }

  fclose(fp);

}

void write_img_info(struct img_set *img_set, char *dbname, char *dirname)
{
  FILE *fp;
  int  k;
  char longname[256];

#if defined(UNIX) || defined(WIN32)
  sprintf(longname, "%s/%s/%s",  dbname, dirname, "db_index");
#else
  sprintf(longname, ":%s:%s:%s", dbname, dirname, "db_index");
#endif

  if ((fp=fopen(longname, "w"))==NULL) {
    fprintf(stderr,"\nError: Cannot open %s\n", longname); 
    exit(1);
  }

  fprintf(fp, "%d\n", img_set->num_img);

  /* write the name of each image in the database */
  for(k=0; k<img_set->num_img; k++) {

    /* get image name */
    fprintf(fp, "%s\n", img_set->img[k].name);
    fprintf(fp, " %d %d",   img_set->img[k].row,     img_set->img[k].col);
    fprintf(fp, " %ld %ld %ld", img_set->img[k].featurept,
								img_set->img[k].featurelen,
								img_set->img[k].thumbpt);
    fprintf(fp, "\n");

  }

  fclose(fp);

}

void read_img_set(struct img_set *img_set, char *dbname,
                  char *subdir,  char *mode)
{
  struct img_info *curimg;
  struct feature *featureblock;
  FILE   *ffeature;
  int    i, status;

  ffeature = open_database(dbname, subdir, mode);
  featureblock = (struct feature *) malloc(img_set->max_img *
                                           sizeof(struct feature));
  img_set->featureblock = featureblock;
  for (i=0; i<img_set->num_img; i++) {
    curimg = &(img_set->img[i]);
    curimg->feature = &featureblock[i];
    status = read_feature(curimg->feature, ffeature,
                          curimg->featurept, curimg->featurelen);
    if (status == FAILED) {
      fprintf(stderr,"Error: can not read img feature %s", curimg->name);
      return;
    }
  }
  for (i=img_set->num_img; i<img_set->max_img; i++) {
    curimg = &(img_set->img[i]);
    curimg->feature = &featureblock[i];
  }
  close_database(ffeature);

}

void allocate_img_set(struct img_set *img_set)
{
  struct feature *featureblock;
  int    i;

  featureblock = (struct feature *) malloc(img_set->max_img *
                                           sizeof(struct feature));

  img_set->featureblock = featureblock;
  for (i=0; i<img_set->max_img; i++) img_set->img[i].feature = &featureblock[i];

}
void write_img_set(struct img_set *img_set, char *dbname,
                   char *subdir,  char *mode)
{
  FILE *ffeature;
  struct img_info *curimg;
  int    i;


  ffeature = open_database(dbname, subdir, mode);

  for (i=0; i<img_set->num_img; i++) {
    curimg = &(img_set->img[i]);

    if (write_feature(curimg->feature, ffeature, &(curimg->featurept), &(curimg->featurelen)) == FAILED) {
      fprintf(stdout,"\nError: Can not write to image database\n");
      return;
    }

  }
  fclose(ffeature);

}


void append_img_info(struct img_set *img_set, struct img_info *img_info)
{
  struct feature *featureblock;
  int i, k;

  if (img_set->num_img == img_set->max_img) {
    if (img_set->max_img >= 100) img_set->max_img *= 2;
    else                         img_set->max_img  = 100;
    img_set->img = (struct img_info *) realloc((void *) img_set->img,
											   img_set->max_img * sizeof(struct img_info));

    if(img_set->img == NULL) {
      printf("\nError: Cannot allocate for heap, size : %d\n",img_set->max_img);
      exit(1);
    }
    for(k=img_set->num_img; k<img_set->max_img; k++) {
      img_set->img[k].ix   = k;
      img_set->img[k].feature  = NULL;
	}

    if (img_set->featureblock != NULL) {
	  featureblock = (struct feature *) realloc((void *) img_set->featureblock,
												img_set->max_img * sizeof(struct feature));
	  img_set->featureblock = featureblock;
	  for (i=0; i<img_set->max_img; i++) img_set->img[i].feature = &featureblock[i];
	}
  }

  /* get the name of each image in the database */

  k = img_set->num_img;

  strcpy(img_set->img[k].name, img_info->name);

  img_set->img[k].ix   = k;
  img_set->img[k].type = 0;
  img_set->img[k].row  = img_info->row;
  img_set->img[k].col  = img_info->col;
  img_set->img[k].num_img = 0;
  img_set->img[k].next = NULL;
  img_set->img[k].child= NULL;
  img_set->img[k].key  = NULL;
  img_set->img[k].nearest = NULL;
  img_set->img[k].featurept = img_info->featurept;
  img_set->img[k].featurelen = img_info->featurelen;
  img_set->img[k].thumbpt = img_info->thumbpt;
  img_set->img[k].searchcount = 0;

  img_set->num_img ++;

}

void get_img_tree(struct img_set *img_set, struct img_set *key_set,
                  char *dbname, char *subdir)
{
  struct img_info *img_info;
  FILE   *fp;
  int    k, ix, root, next, child, key, nearest;
  double dist, size;
  char   longname[256], type[5], type1[5], type2[5], type3[5], type4[5];

#if defined(UNIX) || defined(WIN32)
  sprintf(longname, "%s/%s/%s",  dbname, subdir, "dk_list");
#else
  sprintf(longname, ":%s:%s:%s", dbname, subdir,  "dk_list");
#endif

  if ((fp=fopen(longname, "r"))==NULL) {
    fprintf(stdout,"\nHistogram List Database %s not exist\n", longname); 
    return;
  }

  if (fscanf(fp,"%s %d", type, &root) != 2) {
    fprintf(stderr,"\nError: Corrupted file %s\n", longname);
    exit(1);
  }
  if (strcmp(type, "K") == 0) 
    img_set->root = key_set->root = &(key_set->img[root]);
  else
    img_set->root = key_set->root = &(img_set->img[root]);

  /* get the name of each image in the database */
  for(k=0; k<img_set->num_img+key_set->num_img; k++) {

    /* get image name */
    if (fscanf(fp,"%s %d %s %d %s %d %s %d %s %d %lf %lf ", type, &ix,
	       type1, &next, type2, &child, type3, &key, type4, &nearest,
               &dist, &size) < 1) {
      fprintf(stderr,"\nError: Corrupted file %s\n", longname);
      exit(1);
    }

    if      (strcmp(type, "K") == 0) img_info = &(key_set->img[ix]);
    else                             img_info = &(img_set->img[ix]);

    if      (strcmp(type1, "K") == 0) img_info->next  = &(key_set->img[next]);
    else if (strcmp(type1, "D") == 0) img_info->next  = &(img_set->img[next]);
    else                              img_info->next  = NULL;

    if      (strcmp(type2, "K") == 0) img_info->child = &(key_set->img[child]);
    else if (strcmp(type2, "D") == 0) img_info->child = &(img_set->img[child]);
    else                              img_info->child = NULL;

    if      (strcmp(type3, "K") == 0) img_info->key   = &(key_set->img[key]);
    else if (strcmp(type3, "D") == 0) img_info->key   = &(img_set->img[key]);
    else                              img_info->key   = NULL;

    if      (strcmp(type4, "K") == 0) img_info->nearest  = &(key_set->img[nearest]);
    else if (strcmp(type4, "D") == 0) img_info->nearest  = &(img_set->img[nearest]);
    else                              img_info->nearest  = NULL;

    img_info->dist = dist;
    img_info->size = size;

  }

  fclose(fp);

}

void write_img_tree(struct img_set *img_set, struct img_set *key_set,
                    char *dbname, char *subdir)
{
  struct img_info *img_info;
  FILE   *fp;
  char   longname[256], type[5], type1[5], type2[5], type3[5], type4[5];
  int    k, next, child, key, nearest;
  double dist, size;


#if defined(UNIX) || defined(WIN32)
  sprintf(longname, "%s/%s/%s",  dbname, subdir, "dk_list");
#else
  sprintf(longname, ":%s:%s:%s", dbname, subdir,  "dk_list");
#endif

  if ((fp=fopen(longname, "w"))==NULL) {
    fprintf(stderr,"\nError: Cannot open %s\n", longname); 
    exit(1);
  }

  fprintf(fp, "%s %d\n", "K", img_set->root->ix);

  /* write the name of each image in the database */
  for (k=0; k<img_set->num_img+key_set->num_img; k++) {

    if (k >= img_set->num_img) {
      strcpy(type, "K");
      img_info = &(key_set->img[k - img_set->num_img]);
    } else {
      strcpy(type, "D");
      img_info = &(img_set->img[k]);
    }

    if   (img_info->next  == NULL)    {
      next  = -1; 
      strcpy(type1, "N");
    } else {
      next = (img_info->next)->ix;
      if ((img_info->next)->type == 0) strcpy(type1, "D");
      else                             strcpy(type1, "K");
    }

    if   (img_info->child == NULL)    {
      child = -1; 
      strcpy(type2, "N");
    } else {
      child= (img_info->child)->ix;
      if ((img_info->child)->type == 0) strcpy(type2, "D");
      else                              strcpy(type2, "K");
    }

    if   (img_info->key == NULL)    {
      key = -1; 
      strcpy(type3, "N");
    } else {
      key= (img_info->key)->ix;
      if ((img_info->key)->type == 0) strcpy(type3, "D");
      else                            strcpy(type3, "K");
    }

    if   (img_info->nearest == NULL)    {
      nearest = -1; 
      strcpy(type4, "N");
    } else {
      nearest = (img_info->nearest)->ix;
      if ((img_info->nearest)->type == 0) strcpy(type4, "D");
      else                             strcpy(type4, "K");
    }

    dist = img_info->dist;
    size = img_info->size;

    /* get image name */
    fprintf(fp, "%s %d %s %d %s %d %s %d %s %d %g %g\n", type, img_info->ix, 
            type1, next, type2, child, type3, key, type4, nearest, dist, size);
    fflush(fp);

  }

  fclose(fp);

}

void read_info_pym(struct img_set  *img_set,  struct img_set *key_set,
                   struct info_pym *info_pym, char *dbname, char *subdir)
{
  struct info_level *level;
  struct img_info   *img_info;
  FILE   *fp;
  char   longname[256], type[5];
  int    k, l, x, y, ix;



#if defined(UNIX) || defined(WIN32)
  sprintf(longname, "%s/%s/%s",  dbname, subdir, "dk_pym");
#else
  sprintf(longname, ":%s:%s:%s", dbname, subdir, "dk_pym");
#endif

  if ((fp=fopen(longname, "r"))==NULL) {
    fprintf(stderr,"\nError: Cannot open %s\n", longname); 
    exit(1);
  }

  fscanf(fp, "%d\n", &info_pym->num_level);
  /* write the name of each image in the database */
  for (l=0; l<info_pym->num_level && l < MAXLEVEL; l++) {
    level = &(info_pym->level[l]); 
    level->level = l;
    fscanf(fp, "%d %d\n", &(level->row), &(level->col));
/*
    level->info  = (struct img_info ***) 
                  get_img(level->row, level->col, sizeof(struct img_info *));
    for (y=0; y<level->row; y++) {
      for (x=0; x<level->col; x++) {
        level->info[y][x] = NULL;
      }
    }
*/
  }

  for(k=0; k<img_set->num_img+key_set->num_img; k++) {

    /* get image name */
    if (fscanf(fp,"%s %d %d %d %d", type, &ix, &l, &y, &x) < 1) {
      fprintf(stderr,"\nError: Corrupted file %s\n", longname);
      exit(1);
    }

    if      (strcmp(type, "K") == 0) img_info=&(key_set->img[ix]);
    else if (strcmp(type, "D") == 0) img_info=&(img_set->img[ix]);
    else                             img_info=NULL;
    if (img_info != NULL) {
      img_info->level = l;
      img_info->x     = x;
      img_info->y     = y;
    }
/*
    if (l < MAXLEVEL && l < info_pym->num_level) {
      level = &(info_pym->level[l]);
      level->info[y][x] = img_info;
    }
*/
  }

  fclose(fp);

}

void write_info_pym(struct img_set  *img_set,  struct img_set *key_set,
                    struct info_pym *info_pym, char *dbname, char *subdir)
{
  struct img_info *img_info;
  struct info_level *level;
  FILE   *fp;
  char   longname[256], type[5];
  int    k, l, x, y, ix;


#if defined(UNIX) || defined(WIN32)
  sprintf(longname, "%s/%s/%s",  dbname, subdir, "dk_pym");
#else
  sprintf(longname, ":%s:%s:%s", dbname, subdir, "dk_pym");
#endif


  if ((fp=fopen(longname, "w"))==NULL) {
    fprintf(stderr,"\nError: Cannot open %s\n", longname); 
    exit(1);
  }

  fprintf(fp, "%d\n", info_pym->num_level);

  /* write the name of each image in the database */
  for (l=0; l < info_pym->num_level; l++) {
    if (l >= MAXLEVEL) continue;
    level = &(info_pym->level[l]);
    fprintf(fp, "%d %d\n", level->row, level->col);
/*
    for (i=0; i<level->row; i++) {
      for (j=0; j<level->col; j++) {
        img_info = level->info[i][j];
        if   (img_info != NULL)    {
          img_info->level = l;
          img_info->x     = j;
          img_info->y     = i;
        }
      }
    }
*/
  }

  for (k=0; k<img_set->num_img+key_set->num_img; k++) {

    if (k >= img_set->num_img) {
      strcpy(type, "K");
      img_info = &(key_set->img[k - img_set->num_img]);
    } else {
      strcpy(type, "D");
      img_info = &(img_set->img[k]);
    }

    ix = img_info->ix;
    if (img_info->type == 0) strcpy(type, "D");
    else                     strcpy(type, "K");
    l  = img_info->level;
    x  = img_info->x;
    y  = img_info->y;
    fprintf(fp, "%s %d %d %d %d \n", type, ix, l, y, x);
  }
  fclose(fp);

}

