#include "defs.h"

#define DEBUG			0
#define COMBINATIONNO	8
#define PI				3.141592653589793



static int  free_mapping(struct img_info *img_list);
static int  constrain_mapping(struct img_info *curkey);
static void interdistance(struct img_info **parent, struct img_info **cousin, 
 						  double **disttbl);
static void intradistance(struct img_info **cousin, double **disttbl);
static void Lmap(struct img_info **cousin, double *Ltbl);
static void Amap(struct img_info **cousin, double *Atbl);
static void Bmap(struct img_info **cousin, double *Btbl);
static void huemap(struct img_info **cousin, double *huetbl);
static void texturemap(struct img_info **cousin, double *texttbl);
static double costfunc(double **disttbl1, double **disttbl2, 
					   double *huetbl,    double *texttbl,  
					   int    *num_img1,  int    *num_img2, int *permute);




struct img_info **infoqueue;
int    head, tail;

/*
     |       |
     |   3   |
     |       |
-----+---+---+-----
     | 2 | 3 |
  0  +---+---+  2
     | 0 | 1 |
-----+---+---+------
     |       |
     |   1   |
     |       |
*/
static int permute[24][4]=
  {0, 1, 2, 3, 0, 1, 3, 2, 0, 2, 1, 3, 0, 2, 3, 1, 0, 3, 1, 2, 0, 3, 2, 1,
   1, 0, 2, 3, 1, 0, 3, 2, 1, 2, 0, 3, 1, 2, 3, 0, 1, 3, 0, 2, 1, 3, 2, 0,
   2, 0, 1, 3, 2, 0, 3, 1, 2, 1, 0, 3, 2, 1, 3, 0, 2, 3, 0, 1, 2, 3, 1, 0,
   3, 0, 1, 2, 3, 0, 2, 1, 3, 1, 0, 2, 3, 1, 2, 0, 3, 2, 0, 1, 3, 2, 1, 0};
static double lpos[4][2]={ 0, 0, 1, 0, 0, 1, 1, 1};
static double ppos[4][2]={-1.5, 0.5, 0.5,-1.5, 2.5, 0.5, 0.5, 2.5};

int mapping(struct img_set *img_set) 
{
  struct node_set best_ns;
  struct img_info *curimg;
  struct info_pym info_pym;
  int    i, more;



  /* code from proximity.c */
  infoqueue = (struct img_info **) malloc(sizeof(struct img_info *) *
                                          img_set->num_img * 2 );
  initialize_node_set(&best_ns);


  /* code from mapping.c */
  if (DEBUG) fprintf(stdout, "Begin Mapping\n");
  info_pym.level[0].level = 0;
  for (i=0; i<MAXLEVEL; i++) 
    info_pym.level[i].row   = info_pym.level[i].col = pow(2.0, (double) i+1);

  if (DEBUG) fprintf(stdout, "Free Mapping\n");
  more = free_mapping(img_set->root);
  if (DEBUG) fprintf(stdout, "Free Mapped %d more\n", more);

  info_pym.num_level = 0;
  while (head > tail) {
    curimg = infoqueue[tail++];
    if (info_pym.num_level < curimg->level + 1) {
      info_pym.num_level = curimg->level + 1;
      if (DEBUG) fprintf(stdout, "Constrain Mapped %d, %d more\n", curimg->level, head-tail+1);
      if (DEBUG) fprintf(stdout, "Constrain Mapping %d\n", info_pym.num_level);
    }
    
    more = constrain_mapping(curimg);
  }

  if (DEBUG) fprintf(stdout, "End Mapping\n"); 


  return(0);

}


static int free_mapping(struct img_info *img_list)
{
  int    i, j, p, pick, more;
  int    num_img1[4], num_img2[4];
  double **disttbl1, **disttbl2, huetbl[4], texttbl[4], Ltbl[4], Atbl[4], Btbl[4];
  double maxcost, mincost, cost;
  struct img_info *parent[4], *cousin[4], *curimg;

  more = 0;
  if (img_list == NULL) return(more);

  disttbl1 = (double **) get_img(4, 4, sizeof(double));
  disttbl2 = (double **) get_img(4, 4, sizeof(double));
  for (i=0; i<4; i++) for (j=0; j<4; j++) disttbl1[i][j] = 0;
  for (i=0; i<4; i++) for (j=0; j<4; j++) disttbl2[i][j] = 0;

  /* Set up array and distance matrix */
  for (i=0; i<4; i++) parent[i] = NULL;

  cousin[0] = img_list;
  for (i=1; i<4; i++) {
    if (cousin[i-1] != NULL) cousin[i] = cousin[i-1]->next;
    else                     cousin[i] = NULL;
  }  
  for (i=0; i<4; i++) if (cousin[i] != NULL && cousin[i]->child != NULL) more++;

  for (i=0; i<4; i++) num_img1[i] = 0;
  for (i=0; i<4; i++) num_img2[i] = 0;
  for (i=0; i<4; i++) {
    if (cousin[i] != NULL) {
      num_img2[i] = cousin[i]->num_img;
    }
  }

  intradistance(cousin, disttbl2);
  Lmap(cousin, Ltbl);
  Amap(cousin, Atbl);
  Bmap(cousin, Btbl);
  huemap(cousin, huetbl);
  texturemap(cousin, texttbl);

  mincost = MAXVALUE;
  maxcost = 0.0;
  pick = 0;
  for (p=0; p<24; p++) {
    cost = costfunc(disttbl1, disttbl2, Atbl, texttbl, 
                    num_img1, num_img2, permute[p]);
    if (cost < mincost) {mincost = cost; pick = p;}
  }
    
  for (i=0; i<4; i++) {
    if (cousin[i] == NULL) continue;
    cousin[i]->level = 0;
    cousin[i]->x     = permute[pick][i]%2;
    cousin[i]->y     = permute[pick][i]/2;
  }

  head = tail = 0;
  for (curimg=img_list; curimg!=NULL; curimg=curimg->next)
    if (curimg->type == 1) infoqueue[head++] = curimg;

  free_img((void **) disttbl1);
  free_img((void **) disttbl2);

  return(more);

}

static int constrain_mapping(struct img_info *curkey)
{
  int    i, p, l, x, y, pick, more;
  int    num_img1[4], num_img2[4];
  double **disttbl1, **disttbl2;
  double Ltbl[4], Atbl[4], Btbl[4], huetbl[4], texttbl[4];
  double mincost, cost;
  struct img_info *curimg, *parent[4], *cousin[4];
  
  more = 0;
  if (curkey == NULL) return(0);
  if (curkey->child == NULL) return(0);

  l = curkey->level;
  x = curkey->x;
  y = curkey->y;

  disttbl1 = (double **) get_img(4, 4, sizeof(double));
  disttbl2 = (double **) get_img(4, 4, sizeof(double));

  /* Set up array and distance matrix */
  for (i=0; i<4; i++) parent[i] = NULL;
  for (i=0; i<head; i++) {
    curimg = infoqueue[i];
    if (curimg->level != curkey->level) continue;
    if      (curimg->x == x-1 && curimg->y == y  ) parent[0] = curimg;
    else if (curimg->x == x   && curimg->y == y-1) parent[1] = curimg;
    else if (curimg->x == x+1 && curimg->y == y  ) parent[2] = curimg;
    else if (curimg->x == x   && curimg->y == y+1) parent[3] = curimg;
  }

  cousin[0] = curkey->child;
  for (i=1; i<4; i++) {
    if (cousin[i-1] != NULL) cousin[i] = cousin[i-1]->next;
    else                     cousin[i] = NULL;
  }  
  for (i=0; i<4; i++) 
  if (cousin[i] != NULL && cousin[i]->child != NULL) more++;

  for (i=0; i<4; i++) num_img2[i] = 0;
  for (i=0; i<4; i++) {
    if (cousin[i] != NULL) {
      num_img2[i] = cousin[i]->num_img;
    }
  }
  for (i=0; i<4; i++) num_img1[i] = 0;
  for (i=0; i<4; i++) {
    if (parent[i] != NULL) {
      num_img1[i] = parent[i]->num_img;
    }
  }

  interdistance(parent, cousin, disttbl1);
  intradistance(cousin, disttbl2);
  Lmap(cousin, Ltbl);
  Amap(cousin, Atbl);
  Bmap(cousin, Btbl);
  huemap(cousin, huetbl);
  texturemap(cousin, texttbl);
 
  mincost = 1e+10;
  pick = 0;
  for (p=0; p<24; p++) {
    cost = costfunc(disttbl1, disttbl2, Atbl, texttbl,
                    num_img1, num_img2, permute[p]);
    if (cost < mincost) {mincost = cost; pick = p;}
  }
  
  for (i=0; i<4; i++) {
    if (cousin[i] == NULL) continue;
    cousin[i]->level = l+1;
    cousin[i]->x     = 2*x+permute[pick][i]%2;
    cousin[i]->y     = 2*y+permute[pick][i]/2;
  }
      
  for (curimg=curkey->child; curimg!=NULL; curimg=curimg->next) 
    if (curimg->type == 1) infoqueue[head++] = curimg;

  free_img((void **) disttbl1);
  free_img((void **) disttbl2);
  return(more);

}


static void interdistance(struct img_info **parent, struct img_info **cousin, 
 		   	  double **disttbl)
{
  int    i, j;
  struct node_set ns;
  

  initialize_node_set(&ns);

  /* compute driving force from upper level */
  for (i=0; i<4; i++) for (j=0; j<4; j++) disttbl[i][j] = 0.0;
  for (i=0; i<4; i++) {
    for (j=0; j<4; j++) {
      if (parent[i] != NULL && cousin[j] != NULL) {
        disttbl[j][i]=direct_search(parent[i], cousin[j], &ns);
      } else if (parent[i] == NULL && cousin[j] == NULL) {
        disttbl[j][i]=0.0;
      } else {
        disttbl[j][i]=18.0;
      }
    }
  }
  free_node_set(&ns);
     

}

static void intradistance(struct img_info **cousin, double **disttbl)
{
  int    i, j;
  struct node_set ns;
  
  initialize_node_set(&ns);

  /* compute driving force from same level */
  for (i=0; i<4; i++) for (j=0; j<4; j++) disttbl[i][j] = 0;
  for (i=0; i<4; i++) {
    for (j=i+1; j<4; j++) {
      if (cousin[i] != NULL && cousin[j] != NULL) {
        disttbl[j][i] = disttbl[i][j] = direct_search(cousin[i],cousin[j],&ns);
      } else if (cousin[i] == NULL && cousin[j] == NULL) {
        disttbl[j][i] = disttbl[i][j] = 0.0;
      } else {
        disttbl[j][i] = disttbl[i][j] = 18.0;
      }
    }
  }

  free_node_set(&ns);
     

}

static void Lmap(struct img_info **cousin, double *Ltbl)
{
  int    i, l;
  double lstar;
  struct feature *feature;

  for (i=0; i<4; i++) {
    if (cousin[i] != NULL) {
      feature = cousin[i]->feature;
      for (l=0, lstar=0; l<HIST0_BINS; l++) lstar += feature->hist0[l]*l;
      Ltbl[i] = lstar / HIST0_BINS * 100.0;
    }
  }

}

static void Amap(struct img_info **cousin, double *Atbl)
{
  int    i, l;
  double astar;
  struct feature   *feature;

  for (i=0; i<4; i++) {
    if (cousin[i] != NULL) {
      feature = cousin[i]->feature;
      for (l=0, astar=0; l<HIST1_BINS; l++) astar += feature->hist1[l]*l;
      Atbl[i] = astar * (200.0 / HIST1_BINS) - 100.0;
    }
  }

}

static void Bmap(struct img_info **cousin, double *Btbl)
{
  int    i, l;
  double bstar;
  struct feature   *feature;

  for (i=0; i<4; i++) {
    if (cousin[i] != NULL) {
      feature = cousin[i]->feature;
      for (l=0, bstar=0; l<HIST2_BINS; l++) bstar += feature->hist2[l]*l;
      Btbl[i] = bstar * (200.0 / HIST2_BINS) - 100.0;
    }
  }

}

static void huemap(struct img_info **cousin, double *huetbl)
{
  int    i, l;
  double astar, bstar;
  struct feature   *feature;

  for (i=0; i<4; i++) {
    if (cousin[i] != NULL) {
      feature = cousin[i]->feature;
      for (l=0, astar=0; l<HIST1_BINS; l++) astar += feature->hist1[l]*l;
      for (l=0, bstar=0; l<HIST2_BINS; l++) bstar += feature->hist2[l]*l;
      huetbl[i] = atan2(astar, bstar) / (2. * PI);
    }
  }

}

static void texturemap(struct img_info **cousin, double *texttbl)
{
  int    i, l;
  double Lvalue, avalue, bvalue;
  struct feature   *feature;

  for (i=0; i<4; i++) {
    if (cousin[i] != NULL) {
      feature = cousin[i]->feature;
      for (l=0, Lvalue=0; l<HIST3_BINS; l++) Lvalue += feature->hist3[l]*l;
      for (l=0, avalue=0; l<HIST4_BINS; l++) avalue += feature->hist4[l]*l;
      for (l=0, bvalue=0; l<HIST5_BINS; l++) bvalue += feature->hist5[l]*l;
      Lvalue = Lvalue * (200.0 / HIST3_BINS);
      avalue = avalue * (200.0 / HIST3_BINS);
      bvalue = bvalue * (200.0 / HIST3_BINS);
      texttbl[i] = (Lvalue + avalue + bvalue) / 3.0;
    }
  }

}


static double costfunc(double **disttbl1, double **disttbl2, 
                       double *huetbl,    double *texttbl,  
                       int    *num_img1,  int    *num_img2,	int *permute)
{
  int    i, j, maxnum, minnum, num_img;
  double dx, dy, tx, ty;
  double cost1, cost2, cost3, cost4, cost;
  
  num_img = 0;
  maxnum = 0; 
  minnum = 1000000000;
  for (i=0; i<4; i++)  {
    num_img += num_img2[i];
    if (maxnum < num_img2[i]) maxnum = num_img2[i];
    if (minnum > num_img2[i] && num_img2[i] > 0) minnum = num_img2[i];
  }
  cost = cost1 = cost2 = cost3 = cost4 = 0;
  for (i=0; i<4; i++) { 
    dx = lpos[permute[i]][0];
    dy = lpos[permute[i]][1];

    /* Cost to parent nodes */ 
    for (j=0; j<4; j++) { 
      tx = ppos[j][0];
      ty = ppos[j][1];
      cost1 += num_img1[j] * num_img2[i] *
               (fabs(dx-tx)+fabs(dy-ty)) / (disttbl1[i][j] + 0.001);
    }

    /* Cost to neightbor node */
    for (j=0; j<4; j++) {
      if (i == j) continue;
      tx = lpos[permute[j]][0];
      ty = lpos[permute[j]][1];
      cost2 += num_img2[j] * num_img2[i] *
               (fabs(dx-tx)+fabs(dy-ty)) / (disttbl2[i][j] + 0.001);
    }
    /* Cost due to hue and texture */
    cost3 += num_img * num_img * huetbl[i]  * dy * 0.0001; 
    cost4 += num_img * num_img * texttbl[i] * dx * 0.0001;
     
  }

  cost = cost1 + cost2 + cost3 + cost4;

  return(cost);

}


