#include "defs.h"

#define DEBUG		0
#define KEYNO		10
#define KEYMAX		10
#define KEYSIZE		1.0
#define ITERATION	2
#define PI		3.141592653589793


/* routine for VQ tree partition */
struct img_info *partition(struct img_set  *img_set, struct img_set *key_set,  
                           struct img_info *img_list, int level);
static void pick_key(struct img_info *img_list, struct img_info **key_list, 
					int num_img); 
static void clustering(struct img_set *img_set,   struct img_set *key_set,   
						struct img_info *img_list, struct img_info *key_list,
						int    num_to_find);
static void replace_key (struct img_info *curkey,  struct img_info **img_list,
						struct img_set  *key_set, struct img_info **new_list);

static void mean_key(struct img_info *curkey, struct img_info *img_list);
static void update_distance(struct img_info *curkey);




struct img_info *partition(struct img_set *img_set, struct img_set *key_set, 
						   struct img_info *img_list, int level)
{
  struct img_info *key_list, *new_list, *curkey;
  struct img_info *curimg, *nextimg;
  struct img_info *keyinfo;
  struct feature  *featureblock;
  int    i, j, num_img;
  double adjust;

  keyinfo = (struct img_info *) malloc(sizeof(struct img_info) * KEYNO);
  featureblock = (struct feature *) malloc(sizeof(struct feature) * KEYNO);
  for (curimg=img_list,num_img=0; curimg!=NULL; curimg=curimg->next) num_img++;
  if (DEBUG) fprintf(stdout, "Partitioning %d images\n", num_img);

  for (i=0; i<KEYNO; i++) {
    if (i == KEYNO-1) keyinfo[i].next  = NULL;
    else              keyinfo[i].next  = &(keyinfo[i+1]);
    keyinfo[i].ix    = i;
    keyinfo[i].child = NULL;
    keyinfo[i].key   = NULL;
    keyinfo[i].feature = &featureblock[i];
  }
  key_list = &(keyinfo[0]);

  if (DEBUG) fprintf(stdout, "Pick up keys\n");
  pick_key(img_list, &key_list, num_img);

  if (DEBUG) fprintf(stdout, "Cluster keys\n");
  for (i=0; i<ITERATION; i++) {

    clustering(img_set, key_set, img_list, key_list, num_img);

    adjust = 0;
    for (curkey=key_list; curkey!=NULL; curkey=curkey->next) mean_key(curkey, img_list);

  }

  if (DEBUG) fprintf(stdout, "replace keys\n");
  new_list = NULL;
  for (curkey=key_list; curkey!=NULL; curkey=curkey->next) {
    replace_key(curkey, &img_list, key_set, &new_list);
  }     
  key_list = new_list;

  /* Really partition the image list */
  curimg = img_list;
  while (curimg != NULL) {
    nextimg = curimg->next;
    curkey  = curimg->key;
    curimg->next  = curkey->child;
    curkey->child = curimg;
    curkey->num_img++;
    curimg = nextimg;
  }

  /* Calculate the distance to every images */
  if (DEBUG) fprintf(stdout, "update distance\n");
  for (curkey=key_list; curkey!=NULL; curkey=curkey->next) {
    /* Final update distance matrix without repartition */
    if (curkey->type != 0) update_distance (curkey);
  }     

  /* Recursive partition children */
  curkey = key_list;
  while (curkey != NULL) {
    num_img = 0;
    for (curimg=curkey->child; curimg != NULL; curimg=curimg->next) num_img++;
    if (DEBUG) {
      for (j=0; j<level; j++) fprintf(stdout, "  ");
      if (curkey->type == 1) 
        fprintf(stdout, "Key %d : %d images, Max distance is %g\n", 
                        curkey->ix, num_img, curkey->size);
      else
        fprintf(stdout, "Image %d distance to Key %d is %g\n", 
                        curkey->ix, curkey->key->ix, curkey->dist);
    }
    if (num_img > KEYNO) {
      /* recursive partition */
      curkey->child = partition(img_set, key_set, curkey->child, level+1);
    }
    curkey = curkey->next;
  }  

  free((void *) keyinfo);
  free((void *) featureblock);

  return(key_list);

}

static void pick_key(struct img_info *img_list, struct img_info **key_list,int num_img)
{
  int    keycount, i, j, k, l;
  int    keyindex[KEYNO];
  struct img_info *curimg, *curkey, *prvkey;


  for (i=0; i<KEYNO && i<num_img; ) {
    j = num_img * ((double) rand()) / ((double) RAND_MAX);
    if (j >= num_img) continue;
    for (k=0; k<i; k++) if (keyindex[k] >= j) break;
    if (keyindex[k] > j || k >= i) {
      for (l=KEYNO-1; l>k; l--) keyindex[l]=keyindex[l-1];
      keyindex[k]=j;
      i++;
    }
  }
 
  keycount = 0;
  i = 0;
  curkey = *key_list;
  prvkey = NULL;

  for (curimg=img_list; curimg != NULL && curkey != NULL; curimg=curimg->next) {
    if (keyindex[keycount] == i) {
      /* get the random selected feature */

      copy_img_info(curkey, curimg);
      strcpy(curkey->name, "KEY");
      curkey->type  = 1;
      curkey->key   = curimg->key;
      curkey->nearest  = NULL;
      curkey->dist  = 1e+10;
      curkey->size  = 0;

      /* Initial pyramid */

      copy_feature(curkey->feature, curimg->feature);

      prvkey = curkey;
      curkey = curkey->next;
      keycount ++;
    } 
    curimg->dist = 1e+10;
    i++;
  }

  if (prvkey != NULL) prvkey->next = NULL;
  else                *key_list    = NULL;

}



static void clustering(struct img_set *img_set,   struct img_set  *key_set,  
						struct img_info *img_list, struct img_info *key_list,
						int    num_to_find)
{
  struct img_info *curimg, *curkey, *keytbl[KEYNO];
  struct node     p[KEYNO], pp;
  struct node_set *ns, nstmp;
  double pdist[KEYNO], bestdist, dist;
  int    i, j, k, bestix, maxgroup, keyno;

  if (img_list == NULL) return;

  ns = (struct node_set *) malloc(sizeof(struct node_set) * KEYNO);

  /* Initialize set of location nodes and associated heap */
  for (j=0; j<KEYNO; j++) keytbl[j]=NULL;
  for (curkey=key_list, j=0; curkey!=NULL&&j<KEYNO; curkey=curkey->next, j++) {
    curkey->child = NULL;
    curkey->size  = 0;
    curkey->num_img = 0;
    keytbl[j]=curkey;
    initialize_node_set(&(ns[j]));
  }
  keyno = j;
  initialize_node_set(&nstmp);

  for (curimg=img_list; curimg!=NULL; curimg=curimg->next) {
    curimg->key  = NULL;
    curimg->dist = 1e+10;
  }

  /* Calculate distance from key to images */
  for (curimg=img_list; curimg!=NULL; curimg=curimg->next) {
    for (j=0; j<keyno; j++) {
      curkey = keytbl[j];
      direct_search(curimg, curkey, &nstmp);
      if (get_best_from_node_set(&nstmp, &pp, &dist) == -1) continue;
      insert_into_node_set(&ns[j], &pp, dist);
    }
  }

  for (j=0; j<keyno; j++) {
    if (keytbl[j] == NULL) continue; 
    if (get_best_from_node_set(&(ns[j]),&(p[j]),&(pdist[j]))==-1) {
       fprintf(stdout, "Error(1) while retrieving %d\n", keytbl[j]->ix);
       keytbl[j]=NULL;
    }
  }

  maxgroup = ceil((double) num_to_find / (double) keyno);
  maxgroup = max(KEYMAX, maxgroup);


  for (i=0; i<num_to_find; i++) {
  
    /* Fetch unassigned nearest images for each key */
    for (j=0; j<keyno; j++) {
      if (keytbl[j] == NULL) continue; 
      while (img_set->img[p[j].k].key != NULL && keytbl[j] !=NULL) {
        if (get_best_from_node_set(&(ns[j]),&(p[j]),&(pdist[j]))==-1) {
           fprintf(stdout, "Error(2) while retrieving %d %d\n", 
                            keytbl[j]->ix, keytbl[j]->num_img);
           keytbl[j]=NULL;
        }
      }
    }

    /* Find the nearest images among all key */
    for (j=0, bestix=-1, bestdist=1e+10; j<keyno; j++) 
      if (bestdist>pdist[j] && keytbl[j]!=NULL) {bestix=j; bestdist=pdist[j];}

    if (bestix == -1) {
      fprintf(stdout, "Error while find membership\n");
      for (k=0; k<keyno; k++) 
        fprintf(stdout, "KEY %p Distance %g\n", keytbl[k], pdist[k]);
      exit(0);

    } else if (pdist[bestix] >= 1e+10) {
      fprintf(stdout, "Strange distance between KEY %d and IMAGE %d\n", 
                      keytbl[bestix]->ix, p[bestix].k);
      exit(0);

    } else {
      /* Assign membership */
      img_set->img[p[bestix].k].dist = pdist[bestix];
      img_set->img[p[bestix].k].key  = keytbl[bestix];
      keytbl[bestix]->num_img++;
      /* If over limit, empty the heap */
      if (keytbl[bestix]->num_img >= maxgroup) {
        free_node_set(&(ns[bestix]));
        keytbl[bestix] = NULL;
      }

    }

  }

  for (i=0; i<keyno; i++) 
    if (keytbl[i] != NULL) free_node_set(&(ns[i]));
  free((void *) ns);
  free_node_set(&nstmp);

}

/* Reset key to mean of children */
static void mean_key(struct img_info *curkey, struct img_info *img_list)
{
  struct img_info *curimg;

  reset_feature(curkey->feature);

  for (curimg=img_list; curimg != NULL; curimg=curimg->next) {
    if (curkey == curimg->key) {
      
      add_feature(curkey->feature, curimg->feature, 1.0);

    }
  }  

  curkey->dist = 1e+10;
  curkey->size = 0;

}

/* Replace key with real key and replace key with single child with image*/
static void replace_key(struct img_info *curkey,  struct img_info **img_list,
		   struct img_set  *key_set, struct img_info **new_list)
{
  struct img_info *curimg, *prvimg, *nextimg;
  struct img_info *newkey;
  int    num_img;

  /* Find nearest image */
  num_img = curkey->num_img;

  if (num_img == 0) return;

  if (num_img == 1) {
    curimg = *img_list;
    prvimg = NULL;
    while (curimg != NULL) {
      nextimg = curimg->next;
      if (curkey == curimg->key) {
        if (prvimg == NULL) *img_list     = nextimg;
        else                prvimg->next  = nextimg;

        curimg->next  = *new_list;
        *new_list     = curimg;

        curimg->key   = curkey->key;
        curimg->child = NULL;
        curimg->nearest  = NULL;
        curimg->size  = 0;
        curimg->dist  = curkey->dist;
      } else {
        prvimg = curimg;
      }
      curimg = nextimg;
    }
    return;
  } else {

    /* Replace Key */

    append_img_info(key_set, curkey);
    newkey = &(key_set->img[key_set->num_img-1]);

    newkey->type  = 1;
    newkey->next  = *new_list;
    newkey->child = NULL;
    *new_list     = newkey;

    newkey->key   = curkey->key;
    newkey->nearest  = curkey->nearest;

    if (newkey->feature == NULL)
      newkey->feature = (struct feature *) malloc(sizeof(struct feature));
    copy_feature(newkey->feature, curkey->feature);

    for (curimg=*img_list; curimg != NULL; curimg=curimg->next) {
      if (curkey == curimg->key) curimg->key = newkey;
    }  
  
  }

}

/* Update distance from key to children and key to parent */
static void update_distance(struct img_info *curkey)
{
  struct img_info *curimg, *parkey;
  struct node_set ns;
  double dist, neardist;


  initialize_node_set(&ns);

  curkey->size = 0;
  curkey->dist = MAXVALUE;
  if (curkey->type == 0) return;

  /* update key's distance and size */

  neardist = MAXVALUE;
  curimg = curkey->child;
  while (curimg != NULL) {

	dist = direct_search(curimg, curkey, &ns);

    if (dist < neardist) {
      neardist = dist;
      curkey->nearest = curimg;
    }

    curimg = curimg->next;
  }  

  /* Calculate the distance to parent key */
  parkey = curkey->key;
  if (parkey != NULL) curkey->dist = direct_search(parkey, curkey, &ns);

  free_node_set(&ns);


}


