#include "defs.h"

#define	DEBUG		0
#define	nbmax		64
#define MAX_SUBSET	1000

static double optimize(struct img_set *img_set, struct img_set *key_set,
					   int nbtest, int *tpl_ix, int kth);


int prunebylist(struct img_set *img_set, struct img_set *sub_set, 
				int *submap, int nbtpl, int *tpl_ix, int kth) 
{
  struct img_set  key_set;
  struct node_set best_ns, tmpns;
  struct img_info *curkey, *curimg, *chkimg;
  struct feature  *featureblock;
  struct node p;
  double dist, maxdist, curdist;
  int    i, j;


  if (nbtpl == 0) return(0);

  allocate_img_info(&key_set, nbmax);
  featureblock = (struct feature *) malloc(nbmax * sizeof(struct feature));
  for (i=0; i<nbmax; i++) key_set.img[i].feature = &featureblock[i];

  initialize_node_set(&best_ns);
  initialize_node_set(&tmpns);

  maxdist = optimize(img_set, &key_set, nbtpl, tpl_ix, kth);

  /* Pruning the database using new weighting */
  if (DEBUG) fprintf(stdout, "Pruning the database\n");
  reset_node_set(&best_ns);
  for (j=0; j<img_set->num_img; j++) {
    curimg = &(img_set->img[j]);
    curdist = MAXVALUE;
    for (i=0; i<key_set.num_img; i++) {
      curkey = &key_set.img[i];
      dist = direct_search(curimg, curkey, &tmpns);
      if (dist < curdist) curdist = dist;
    }
    p.k = j;
    p.key = 0;
    insert_into_node_set(&best_ns, &p, curdist);
  }

  /* maxdist *= num_to_find / 100.0; */
  sub_set->num_img = 0;
  for(i=0; i<img_set->num_img; i++) {
    if (i >= MAX_SUBSET) break; 
    if (get_best_from_node_set(&best_ns,&p,&dist) == -1) break;
    if (dist > maxdist) break; 

    curimg = &img_set->img[p.k];
    append_img_info(sub_set, curimg);
    chkimg = &sub_set->img[sub_set->num_img-1];
    copy_feature(chkimg->feature, curimg->feature);
    submap[sub_set->num_img-1] = p.k;

  }
/*
  for(; i<img_set->num_img && dist < maxdist; i++) {
    if (get_best_from_node_set(&best_ns,&p,&dist) == -1) break;
    if (dist > maxdist) break; 
  }
  fprintf(stdout, " %d\n", i);
*/

  /* free set of location nodes for each imgs */
  free_node_set(&tmpns);
  free_node_set(&best_ns);
  free(featureblock);
  free_img_info(&key_set);

  return(0);

}

/* Use all relevant imgs as template */
static double optimize(struct img_set *img_set, struct img_set *key_set, int nbtest, int *tpl_ix, int kth)
{
  struct img_info *curkey, *curimg;
  struct node_set  tmpns;
  double dist, curdist, maxdist, disttbl[nbmax];
  int    i, j;

  initialize_node_set(&tmpns);

  /* Assign every img to one cluster */
  key_set->num_img = 0;
  for (i=0; i<nbtest; i++) {
    curimg = &(img_set->img[tpl_ix[i]]);
    append_img_info(key_set, curimg);
    curkey = &(key_set->img[key_set->num_img-1]);
    curkey->type = 1;
    copy_feature(curkey->feature, curimg->feature);
  }

  for (i=0; i<nbtest; i++) {
    curimg = &(img_set->img[tpl_ix[i]]);
    curdist = MAXVALUE;
    for (j=0; j<nbtest; j++) {
      if (i == j && nbtest > 1) continue;
      curkey = &(img_set->img[tpl_ix[j]]);
      dist = direct_search(curimg, curkey, &tmpns);
      if (dist < curdist) curdist = dist;
    }
    disttbl[i] = curdist;
  }

  for (i=0; i<nbtest-1; i++) {
    for (j=i+1; j<nbtest; j++) {
      if (disttbl[i] > disttbl[j]) {
        dist = disttbl[i];
        disttbl[i] = disttbl[j];
        disttbl[j] = dist;
      }
    }
  }

  maxdist = 0;
  if      (nbtest == 1)  maxdist = 1e+10;
  else if (kth <= 0)     maxdist = 0.0;
  else if (kth > nbtest) maxdist = 18.0;
  else                   maxdist = disttbl[kth-1];

  return(maxdist);

}


