#include "defs.h"

#define DEBUG			0
#define COMBINATIONNO	8
#define PI				3.141592653589793

static double mean_key(struct img_info *curkey);
static double update_distance(struct img_info *curkey);
static double calculate_distance(struct img_info *curkey, struct img_info *curimg, struct node_set *ns);
static void update_nearest(struct img_info *curkey, int nbrel, int *relevance);


int combination[COMBINATIONNO][5] = { 3,  4,  5,  6, 4,
                                      7,  8,  4,  2, 4,
                                      3,  9, 10,  2, 4,
                                      1, 11, 12,  6, 4,
                                      1,  5, 13, 14, 4,
                                      3,  4,  2, -1, 3,
                                      1,  5,  6, -1, 3,
                                      1,  2, -1, -1, 2};




void *regroup(struct img_set *img_set, struct img_set *key_set, 
	      	  struct img_set *new_set, struct img_info *root,
              int    level, int nbrel, int *relevance)
{
  struct img_info *curkey, *newkey, *prvkey, *chkkey;
  struct img_info *keytbl[16];
  int    i, j, minix, minbranch;
  double mincost, mincost2, cost, cost2, maxnum, maxlvl;

  if (root == NULL) return(NULL);
  if (root->type == 0) return(NULL);

  keytbl[0] = root;
  for (i=1; i<16; i++) {
    j = (i + 1) / 2 - 1;
    if (keytbl[j] != NULL && keytbl[j]->type == 1) {
      if ((i + 1) % 2 == 0) keytbl[i] =  keytbl[j]->child;
      else                  keytbl[i] = (keytbl[j]->child)->next;
    } else {
      keytbl[i] = NULL;
    }
  }
  
  /* Compute cost for all permuation */
  minbranch = 0;
  mincost  = MAXVALUE;
  mincost2 = MAXVALUE;
  minix = 0;
  for (i=0; i<COMBINATIONNO; i++) {
    /* Try to minimize subtree size */
    maxnum = maxlvl = 0;
    for (j=0; j<4 && combination[i][j] >= 0; j++) {
      curkey = keytbl[combination[i][j]];
      if      (curkey == NULL) {
        maxnum = MAXVALUE;
        maxlvl = MAXVALUE;
        break;
      } else {
        if (maxnum < curkey->num_img) maxnum = curkey->num_img;
        if (maxlvl < curkey->depth)   maxlvl = curkey->depth;
      }
    }
    cost2 = cost = maxnum;
/*
    if (combination[i][4] < minbranch) continue;
    if (cost < MAXVALUE) minbranch = combination[i][4];
*/
    if (mincost > cost || (mincost == cost && mincost2 > cost2)) { 
      mincost = cost; 
      mincost2 = cost2; 
      minix = i;
/*    Mask this if force binary tree direct collapse to quad-tree */
/*
      if (maxnum < root->num_img) break;
*/
    }
  }


  /* Merge binary tree children to quadtree (maybe generic tree later) */
  /* Only work for 8 children */
  prvkey = NULL;
  for (i=0; i<4 && combination[minix][i] >= 0; i++) {
    curkey = keytbl[combination[minix][i]];
    if (curkey == NULL) {
      fprintf(stdout, "Strange! root %d:%d %g:%g\n", root->type, root->ix, mincost, mincost2);
      exit(0);
    }

    if (curkey->type == 1) {
      append_img_info(new_set, curkey);
      newkey = &(new_set->img[new_set->num_img-1]);
      strcpy(newkey->name, "KEY");
      newkey->type  = 1;
      newkey->child = curkey->child;
      newkey->num_img = curkey->num_img;
      newkey->depth = curkey->depth;
      for (chkkey = curkey->child; chkkey != NULL; chkkey = chkkey->next) 
        chkkey->key = newkey; 
    } else {
      newkey = curkey;
    }

    newkey->key  = root;
    newkey->next = NULL;
    if (prvkey == NULL) root->child  = newkey;
    else                prvkey->next = newkey;

    prvkey = newkey;
  }
  if (prvkey != NULL) prvkey->next = NULL;

  /* Recursive regroup children */
  curkey = root->child;
  while (curkey != NULL) {
    if (DEBUG) {
      for (j=0; j<level; j++) fprintf(stdout, "  ");
      if (curkey->type == 1) 
        fprintf(stdout, "Key %d : %d images, Max distance is %g\n", 
                        curkey->ix, curkey->num_img, curkey->size);
      else
        fprintf(stdout, "Image %d \n", curkey->ix);
    }

    if (curkey->type == 1) {
      /* recursive regroup */

      if (curkey->feature == NULL)
		curkey->feature = (struct feature *) malloc(sizeof(struct feature));
      reset_feature(curkey->feature);
      regroup(img_set, key_set, new_set, curkey, level+1, nbrel, relevance);
      mean_key(curkey);  
      update_distance(curkey); 
      update_nearest(curkey, nbrel, relevance); 
  
    }

    curkey = curkey->next;
  }  
  return(curkey);

}


/* Reset key to mean of children */
static double mean_key(struct img_info *curkey)
{
  struct img_info *curimg;

  if (curkey->type == 0) return(0.0);
  if (curkey->num_img == 0) return(0.0);

  reset_feature(curkey->feature);
  curimg = curkey->child;
  while (curimg != NULL) {
    if (curimg->type == 1) mean_key(curimg);
    add_feature(curkey->feature, curimg->feature, 1.0);

    curimg = curimg->next;
  }  

  curkey->dist = MAXVALUE;
  curkey->size = 0;
  return(curkey->dist);

}


/* Update distance from key to children and key to parent */
static double update_distance(struct img_info *curkey)
{
  struct img_info *curimg, *parkey;
  struct node_set ns;
  double dist;

  curkey->size = 0;
  curkey->dist = MAXVALUE;

  if (curkey->type == 0) return(0.0);
  if (curkey->num_img == 0) return(0.0);

  initialize_node_set(&ns);


  /* update key's distance and size */
  curkey->size = 0;

  curimg = curkey->child;
  while (curimg != NULL) {

    dist = calculate_distance(curkey, curimg, &ns);
    if (curkey->size < dist) curkey->size = dist;

    curimg = curimg->next;
  }  

  /* Calculate the distance to parent key */
  parkey=curkey->key;
  if (parkey != NULL) curkey->dist = direct_search(parkey, curkey, &ns);

  free_node_set(&ns);
  return(curkey->dist);

}


/* Update distance from key to children and key to parent */
static double calculate_distance(struct img_info *curkey, struct img_info *curimg, struct node_set *ns)
{
  struct img_info *chkimg;
  double dist, fardist;

  if (curkey->type == 0) return(0.0);
  if (curkey->num_img == 0) return(0.0);

  /* update key's distance and size */
  if (curimg->type == 0) {
    dist = direct_search(curimg, curkey, ns);
    if (curimg->key == curkey) curimg->dist = dist;
	return(dist);
  } else {

    fardist = 0;
    chkimg = curimg->child;
    while (curimg != NULL) {

      dist = direct_search(curkey, chkimg, ns);
      if (dist > fardist) fardist = dist;

      curimg = curimg->next;
	}  
  }

  return(fardist);

}


static void update_nearest(struct img_info *curkey, int nbrel, int *relevance)
{

  struct img_info *childkey, *curimg;
  struct node_set ns;
  double dist, mindist;
  int    i;
  
  if (curkey->type == 0) return;

  /* Redefine the nearest image */
  mindist = MAXVALUE;
  childkey=curkey->child;
  initialize_node_set(&ns);

  for (childkey=curkey->child; childkey!=NULL; childkey=childkey->next) {
    if (childkey->type == 1) curimg = childkey->nearest;
    else		     curimg = childkey;

    dist = direct_search(curimg, curkey, &ns);
    if (dist < mindist) { curkey->nearest = curimg; mindist = dist; }
    for (i=0; i<nbrel; i++) {
      if (relevance[i]==curimg->ix) {
        curkey->nearest = curimg; 
        mindist = dist; 
        for (;childkey->next!=NULL; childkey=childkey->next);
      } 
    }
    
  }

  free_node_set(&ns);

}

