#include "defs.h"
#define	DEBUG		0
#define	szmax		256
#define PRANK		2

/*
#define col(a,b)	(((a) > (b)) ? (a) : (b))
#define row(a,b)	(((a) < (b)) ? (a) : (b))
*/
static double optimize(struct img_set *img_set, struct img_set *key_set,
		        int nbtpl, int *tpl_ix, int *sztest, int *test_ix);
static void estimate(struct img_set *img_set, struct img_set *key_set,
                     float *weight, int nbtpl, int sztest, int *test_ix);
static double distance(float *weight, double *feature);
static double energy(double *disttbl, int sztest, int nbtpl);
static void conjugate(float *weight, double ***feature,
                      int sztest, int nbtpl, int nbcluster);
static void evaluate(float  *weight, double ***feature, 
                     double *fvalue, double *gvalue, 
                     int sztest, int nbtpl, int nbcluster, int debug);
static double solvealpha(float *weight, float *dweight, double ***feature,
                         int sztest, int nbtpl, int nbcluster);

int relevance(struct img_set *img_set, int nbtpl, int *tpl_ix, float *weight) 
{
  struct img_set key_set;
  struct feature  *featureblock;
  double maxdist;
  int    i;
  int    sztest, *test_ix;


  if (nbtpl == 0) return(0);

  allocate_img_info(&key_set, nbtpl);
  featureblock = (struct feature *) malloc(nbtpl * sizeof(struct feature));
  for (i=0; i<nbtpl; i++) key_set.img[i].feature = &featureblock[i];
  
  test_ix = (int *) malloc(szmax * sizeof(int));  


  maxdist = optimize(img_set, &key_set, nbtpl, tpl_ix, &sztest, test_ix);

  /* Apply conjugate gradient  to compute the weighting 
     which best separate the relevant set and the irrelevant set */
  for (i=0; i<nbweight; i++) weight[i] = 0.5;
  if (nbtpl > 1) estimate(img_set, &key_set, weight, nbtpl, sztest, test_ix);

  return(0);
}

/* Use all relevant imgs as template */
static double optimize(struct img_set *img_set, struct img_set *key_set,
						int nbtpl, int *tpl_ix, int *sztest, int *test_ix)
{
  struct img_info *curkey, *curimg;
  struct node_set  tmpns;
  int    i, j, skip;

  initialize_node_set(&tmpns);

  /* Assign every img to one cluster */
  key_set->num_img = 0;
  for (i=0; i<nbtpl; i++) {
    curimg = &(img_set->img[tpl_ix[i]]);
    append_img_info(key_set, curimg);
    curkey = &(key_set->img[key_set->num_img-1]);
    curkey->type = 1;
    copy_feature(curkey->feature, curimg->feature);
  }

 
  for (i=0; i<nbtpl; i++) test_ix[i] = tpl_ix[i];
  *sztest = nbtpl;

  srand(time(NULL));

  if (img_set->num_img > szmax) {
    while (*sztest < szmax && *sztest < img_set->num_img) {
      j = img_set->num_img * ((double) rand()) / ((double) RAND_MAX);
      if (j >= img_set->num_img) continue;
      for (i=0, skip=0; i<*sztest; i++) if (test_ix[i] == j) skip = 1;
      if (skip == 1) continue;

      curimg = &img_set->img[j];
      test_ix[*sztest] = curimg->ix;
      (*sztest)++;
	} 
  } else {
    for (j=0; j<img_set->num_img; j++) {
      for (i=0, skip=0; i<*sztest; i++) if (test_ix[i] == j) skip = 1;
      if (skip == 1) continue;

      curimg = &img_set->img[j];
      test_ix[*sztest] = curimg->ix;
      (*sztest)++;
	}
  }

  return(0.0);

}


/* Apply conjugate gradient  to compute the weighting 
   which best separate the relevant set and the irrelevant set */
static void estimate(struct img_set *img_set, struct img_set *key_set,
                     float *weight, int nbtpl, int sztest, int *test_ix)
{
  int    i, j, k, l;
  double ***feature;
  struct node_set   best_ns;
  struct img_info *curkey, *curimg;

  feature = (double ***) multialloc(sizeof(double), 3, 
                                    sztest, key_set->num_img, nbfeature);
  initialize_node_set(&best_ns);

  /* Compute |xi-yj| */
  for (i=0; i<key_set->num_img; i++) {
    curkey = &(key_set->img[i]);
    for (j=0; j<sztest; j++) {
      curimg = &(img_set->img[test_ix[j]]);
      for (k=0; k<nbfeature; k++) {
        for (l=0; l<nbweight; l++) weight[l] = 0; 
        weight[k] = 1.0;
        setWeight(weight);
        feature[j][i][k] = direct_search(curimg, curkey, &best_ns);
      }
    }
  }

  conjugate(weight, feature, sztest, nbtpl, key_set->num_img);
  
  multifree(feature, 3);

}

static double distance(float *weight, double *feature)
{
  int    i;
  double dist;

  dist = 0;
  for (i=0; i<nbweight; i++) dist += feature[i]*weight[i];

  return(dist);

}

static double energy(double *disttbl, int sztest, int nbtpl)
{
  int    im, j, better[szmax];
  double cumscore;

  for (j=0; j<nbtpl; j++) better[j] = 0;

  cumscore = 0;
  for (j=0; j<nbtpl; j++) {
    for (im=0; im<sztest; im++) {
      if (disttbl[im] <= disttbl[j]) better[j]++;
    }
    cumscore -= log(better[j]);
  } 

  return(cumscore);

}

static void conjugate(float *weight, double ***feature,
                      int sztest, int nbtpl, int nbcluster)
{
  int    i, k, l, allzero;
  float  dweight[nbweight];
  double alpha, beta, a, b, adjust;
  double fvalue, initvalue, endvalue;
  double gvalue[nbweight], newgvalue[nbweight], dgvalue[nbweight];

  /* Step 1: k=0, Select Initial weighting */
  for (i=0; i<nbweight; i++) weight[i] = 0.5;
  if (DEBUG > 1) {
    fprintf(stdout, "\nInitial Weighting");
    for (i=0; i<nbweight; i++) fprintf(stdout, "%g ", weight[i]);
    fprintf(stdout, "\n");
  }


  /* Step 2: Calculate Initalial gradient and dweight */
  evaluate(weight, feature, &initvalue, gvalue, sztest, nbtpl, nbcluster, 0);
  if (DEBUG) fprintf(stdout, "Run %d: fvalue: %g alpha %g\n", 0, initvalue, alpha);
  for (i=0; i<nbweight; i++) dweight[i] = -gvalue[i];

  for (l=0; l<2; l++) {
  if (DEBUG > 1) {
    for (i=0; i<nbweight; i++) fprintf(stdout, "%g ", dweight[i]);
    fprintf(stdout, "\n");
  }

  for (k=0; k<nbfeature; k++) {

    allzero = 1;
    for (i=0; i<nbweight; i++) if (gvalue[i] != 0) allzero = 0;
    if (allzero == 1) break;

    /* Step 3: alpha = argmin f(weight+alpha*dweight) */
    alpha = solvealpha(weight, dweight, feature, sztest, nbtpl, nbcluster);
    if (DEBUG > 1) fprintf(stdout, "Alpha %g\n", alpha);

    /* Step 4: Update weight */
    for (i=0; i<nbweight; i++) weight[i] += alpha * dweight[i];
    for (i=0; i<nbweight; i++) if (weight[i] < 0) weight[i] = 0;
    if (DEBUG > 1) {
      for (i=0; i<nbweight; i++) fprintf(stdout, "%g ", weight[i]);
      fprintf(stdout, "\n");
    }

    /* Step 5: Calculate new gradient */
    evaluate(weight, feature, &fvalue, newgvalue, sztest, nbtpl, nbcluster, 0);
    if (DEBUG) fprintf(stdout, "Run %d: fvalue: %g alpha %g\n", k, fvalue, alpha);
    if (alpha == 0) {
      if (DEBUG) fprintf(stdout, "alpha is 0, reset dweight\n");
      for (i=0; i<nbweight; i++) gvalue[i] = newgvalue[i];
      for (i=0; i<nbweight; i++) dweight[i] = -gvalue[i];
      continue;
    }

    /* Step 6: Calculate beta */
    /* Hestenes-Stiefel Formula */
/*
    for (i=0; i<nbweight; i++) dgvalue[i] = newgvalue[i] - gvalue[i];
    for (i=0, a = 0; i<nbweight; i++) a += newgvalue[i] * dgvalue[i];
    for (i=0, b = 0; i<nbweight; i++) b += dweight[i] * dgvalue[i];
    beta = a / b;
*/

    /* Fletcher-Reeves Formula */
    for (i=0; i<nbweight; i++) dgvalue[i] = newgvalue[i] - gvalue[i];
    for (i=0, a = 0; i<nbweight; i++) a += newgvalue[i] * dgvalue[i];
    for (i=0, b = 0; i<nbweight; i++) b += gvalue[i] * gvalue[i];
    beta = a / b;
/*    if (beta < 0) beta = 0; */

    if (DEBUG > 1) fprintf(stdout, "Beta %g\n", beta);
    for (i=0; i<nbweight; i++) gvalue[i] = newgvalue[i]; /* update gradient */

    /* Step 7: Update dweight */
    for (i=0; i<nbweight; i++) dweight[i] = -gvalue[i] + beta * dweight[i];
    if (DEBUG > 1) {
      for (i=0; i<nbweight; i++) fprintf(stdout, "%g ", dweight[i]);
      fprintf(stdout, "\n");
    }

    /* Step 8: k=k+1, go to step 3 */
  }
/*
    for (i=0; i<nbweight; i++) {
	  adjust =  2.0 * (((double) rand()) / ((double) RAND_MAX) - 0.5);
	  dweight[i] += adjust;
	}
*/
  evaluate(weight, feature, &endvalue, gvalue, sztest, nbtpl, nbcluster, 0);
  }

  if (DEBUG) fprintf(stdout, "%g %g ", initvalue, endvalue);

}


static void evaluate(float  *weight, double ***feature, 
                     double *fvalue, double *gvalue, 
                     int sztest, int nbtpl, int nbcluster, int debug)
{
  int    im, i, j, n, ystar, xplus, xminus, better[szmax];
  double fx, dist, dplus, dminus, disttbl[szmax+2], dfeature[szmax+2][nbweight];
  double sum;

  *fvalue = 0;
  for (i=0; i<nbweight; i++) gvalue[i] = 0;
  for (i=0; i<szmax+2; i++) for (j=0; j<nbweight; j++) dfeature[i][j] = 0;

    for (n=0; n<nbtpl; n++) {
      disttbl[n] = MAXVALUE;
      for (j=0; j<nbcluster;  j++) {
        if (j == n) continue;
        dist = distance(weight, feature[n][j]);
        if (dist < disttbl[n]) {disttbl[n] = dist; ystar = j;}
      }
      for (j=0; j<nbweight; j++) dfeature[n][j] = feature[n][ystar][j];

      for (i=nbtpl; i<sztest; i++) {
        disttbl[i] = MAXVALUE;
        for (j=0; j<nbcluster;  j++) {
          if (j == n) continue;
          dist = distance(weight, feature[i][j]);
          if (dist < disttbl[i]) {disttbl[i] = dist; ystar = j;}
        }
        for (j=0; j<nbweight; j++) dfeature[i][j] = feature[i][ystar][j];
      }
      for (j=0; j<nbweight; j++) dfeature[szmax][j] = 0;
      disttbl[szmax] = distance(weight, dfeature[szmax]);
      for (j=0; j<nbweight; j++) dfeature[szmax+1][j] = 1;
      disttbl[szmax+1] = distance(weight, dfeature[szmax+1]);

      xplus = xminus = -1;
      dplus = dminus = MAXVALUE;
      better[n] = 1;
      for (im=nbtpl; im<sztest; im++) {
        if (im == n) continue;
        if (disttbl[im] < disttbl[n]) {
          better[n]++;
          if (disttbl[n] - disttbl[im] < dminus) {
            xminus = im;
            dminus = disttbl[n] - disttbl[im];
          }
        } else {
          if (disttbl[im] - disttbl[n] < dplus) {
            xplus = im;
            dplus = disttbl[im] - disttbl[n];
          }
        }
      }
      /* Need to think a better way to handle boundary condition */
      if (xminus == -1) xminus = n;
      if (xplus  == -1) xplus  = szmax+1;
      if (DEBUG > debug) 
        fprintf(stdout, "%d : %d:%g;", n, better[n], disttbl[n]);
/*
        fprintf(stdout, "%d : %d:%g : %d,%g : %d,%g\n", 
              n, better[n], disttbl[n], 
                xplus, disttbl[xplus], xminus, disttbl[xminus]);
*/
	  if (fabs(disttbl[xplus]-disttbl[xminus]) > 1e-20) 
      fx = better[n] + 
           (disttbl[n]-disttbl[xminus]) / (disttbl[xplus]-disttbl[xminus]);
	  else
	  fx = better[n];

      *fvalue -= pow(fabs(fx), PRANK);
      if (fabs(disttbl[xplus]-disttbl[xminus]) > 1e-20 && fabs(fx) > 1e-20)
      for (i=0; i<nbweight; i++)
        gvalue[i] -= pow(fabs(fx), PRANK-1) * PRANK *
        ((dfeature[n][i]-dfeature[xminus][i])*(disttbl[xplus]-disttbl[xminus]) -
         (disttbl[n]-disttbl[xminus])*(dfeature[xplus][i]-dfeature[xminus][i]))/
        ((disttbl[xplus]-disttbl[xminus])*(disttbl[xplus]-disttbl[xminus]));
    }

  for (i=0, sum=0.0; i<nbweight; i++) sum += gvalue[i] * gvalue[i];
  if (sum > 1e-20) for (i=0; i<nbweight; i++) gvalue[i] /= sqrt(sum);
 
 
  if (DEBUG > debug) {
    fprintf(stdout, "fvalue %g\n", *fvalue);
/*
    for (i=0; i<nbweight; i++) fprintf(stdout, "%g ", gvalue[i]);
    fprintf(stdout, "\n");
*/
  }

}

int dnumber=0;

/* Use Golden Section Method to solve alpha */
static double solvealpha(float *weight, float *dweight, double ***feature,
                         int sztest, int nbtpl, int nbcluster)
{
  int    j;
  float  newweight[nbweight];
  double lfvalue, rfvalue, ifvalue, newlfvalue, newrfvalue, error, thou;
  double alpha, lalpha, ralpha, newlalpha, newralpha;
  double gvalue[nbweight];
/*
  double **ftable;
  char   matname[80];

  ftable = (double **) get_img(2001, 1, sizeof(double));
  for (i=0; i<=2000; i++) {ftable[0][i] = 0;}
  for (i=0; i<=2000; i++) {
    lalpha = -5.0 + 0.005 * i; 
    for (j=0; j<nbweight; j++) newweight[j]=weight[j]+lalpha*dweight[j];
    evaluate(newweight,feature,&ftable[0][i],gvalue,sztest,nbtpl,nbcluster);
  }
  sprintf(matname, "fvalue%d", dnumber);
  write_matfile(matname, ftable, 1, 2001);
  free_img((void **) ftable);
  dnumber++;
*/

  lalpha = -2.0;
  ralpha = +2.0;
  for (j=0; j<nbweight; j++) {
    if (fabs(dweight[j]) >= 1e-20 && weight[j] >= 1e-20) {
      alpha = - weight[j] / dweight[j];
      if (dweight[j] > 0 && lalpha < alpha) lalpha = alpha;
      if (dweight[j] < 0 && ralpha > alpha) ralpha = alpha;
    }
  }

  for (j=0; j<nbweight; j++) newweight[j]=weight[j];
  evaluate(newweight, feature, &ifvalue, gvalue, sztest, nbtpl, nbcluster, 3);

  /* Initialize Left point and Evaluate f value at lalpha */
  for (j=0; j<nbweight; j++) newweight[j]=max(0.0, weight[j]+lalpha*dweight[j]);
  evaluate(newweight, feature, &lfvalue, gvalue, sztest, nbtpl, nbcluster, 3);

  /* Initialize Right point and Evaluate f value at ralpha */
  for (j=0; j<nbweight; j++) newweight[j]=max(0.0, weight[j]+ralpha*dweight[j]);
  evaluate(newweight, feature, &rfvalue, gvalue, sztest, nbtpl, nbcluster, 3);

  if (DEBUG > 3) fprintf(stdout, "Lvalue %g:%g, Rvalue %g:%g%\n",
                         lalpha, lfvalue, ralpha, rfvalue);

  thou = (3.0 - sqrt(5.0)) / 2.0;
  newlalpha = lalpha +      thou  * (ralpha - lalpha);
  for (j=0; j<nbweight; j++) newweight[j]=max(0.0, weight[j]+newlalpha*dweight[j]);
  evaluate(newweight, feature, &newlfvalue, gvalue, sztest, nbtpl, nbcluster, 3);

  newralpha = lalpha + (1 - thou) * (ralpha - lalpha);
  for (j=0; j<nbweight; j++) newweight[j]=max(0.0, weight[j]+newralpha*dweight[j]);
  evaluate(newweight, feature, &newrfvalue, gvalue, sztest, nbtpl, nbcluster, 3);

  if (newlfvalue < newrfvalue) {
    if (lfvalue < rfvalue || lfvalue < newrfvalue) {
      lalpha  = newlalpha;
      lfvalue = newlfvalue;
    } else{
      ralpha  = newralpha;
      rfvalue = newrfvalue;
    }
  } else {
    if (lfvalue > rfvalue || rfvalue < newlfvalue) {
      ralpha  = newralpha;
      rfvalue = newrfvalue;
    } else {
      lalpha  = newlalpha;
      lfvalue = newlfvalue;
    }
  }

  error = fabs(ralpha - lalpha);
  while (error > 1e-5) {

    if (lalpha == newlalpha) {
      newlalpha  = newralpha;
      newlfvalue = newrfvalue;
      newralpha  = lalpha + (1 - thou) * (ralpha - lalpha);

      /* Compute new Right point and Evaluate f value at ralpha */
      for (j=0; j<nbweight; j++) newweight[j]=max(0.0, weight[j]+newralpha*dweight[j]);
      evaluate(newweight,feature,&newrfvalue,gvalue,sztest,nbtpl,nbcluster, 3);
    } else {
      newralpha  = newlalpha;
      newrfvalue = newlfvalue;
      newlalpha  = lalpha +      thou  * (ralpha - lalpha);

      /* Compute new Left  point and Evaluate f value at ralpha */
      for (j=0; j<nbweight; j++) newweight[j]=max(0.0, weight[j]+newlalpha*dweight[j]);
      evaluate(newweight,feature,&newlfvalue,gvalue,sztest,nbtpl,nbcluster, 3);
    }
    if (newlfvalue < newrfvalue) {
      if (lfvalue < rfvalue || lfvalue < newlfvalue) {
        lalpha  = newlalpha;
        lfvalue = newlfvalue;
      } else{
        ralpha  = newralpha;
        rfvalue = newrfvalue;
      }
    } else {
      if (lfvalue > rfvalue || rfvalue < newrfvalue) {
        ralpha  = newralpha;
        rfvalue = newrfvalue;
      } else {
        lalpha  = newlalpha;
        lfvalue = newlfvalue;
      }
    }
    if (DEBUG > 3) fprintf(stdout, "Lvalue %g:%g, Rvalue %g:%g%\n",
                           lalpha, lfvalue, ralpha, rfvalue);

    error = fabs(ralpha - lalpha);
  }
 
  if (lfvalue > rfvalue) return(lalpha);
  else                   return(ralpha);

}

/* Use Secant Method to solve alpha */
static double solvealpha2(float *weight, float *dweight, double ***feature,
                          int sztest, int nbtpl, int nbcluster)
{
  int    j;
  float  newweight[nbweight];
  double newfvalue, lfvalue, rfvalue, error;
  double lalpha, ralpha, newalpha;
  double lvalue, rvalue, newvalue;
  double gvalue[nbweight];
/*
  double **ftable, **fgrad;
  char   matname[80];

  ftable = (double **) get_img(2001, 1, sizeof(double));
  fgrad  = (double **) get_img(2001, 1, sizeof(double));
  for (i=0; i<=2000; i++) {ftable[0][i] = 0; fgrad[0][i] = 0;}

  for (i=0; i<=2000; i++) {
    lalpha = -1.0 + 0.001 * i; 
    for (j=0; j<nbweight; j++) newweight[j]=weight[j]+lalpha*dweight[j];
    evaluate(newweight,feature,&ftable[0][i],gvalue,sztest,nbtpl,nbcluster, 3);
    for (j=0,fgrad[0][i]=0; j<nbweight; j++) fgrad[0][i]+=gvalue[j]*dweight[j];
  }
  sprintf(matname, "fvalue%d", dnumber);
  write_matfile(matname, ftable, 1, 2001);
  sprintf(matname, "fgrad%d", dnumber);
  write_matfile(matname, fgrad,  1, 2001);
  free_img((void **) ftable);
  free_img((void **) fgrad);
  dnumber++;
*/
  /* Initialize Left point */
  lalpha =-1.0;

  /* Evaluate f derivative w.r.t. alpha */
  for (j=0; j<nbweight; j++) newweight[j]=weight[j]+lalpha*dweight[j];
  evaluate(newweight, feature, &lfvalue, gvalue, sztest, nbtpl, nbcluster, 3);
  for (j=0,lvalue=0; j<nbweight; j++) lvalue += gvalue[j] * dweight[j];

  /* Initialize Right point */
  ralpha = 1.0;

  /* Evaluate f derivative w.r.t. alpha */
  for (j=0; j<nbweight; j++) newweight[j]=weight[j]+ralpha*dweight[j];
  evaluate(newweight, feature, &rfvalue, gvalue, sztest, nbtpl, nbcluster, 3);
  for (j=0,rvalue=0; j<nbweight; j++) rvalue += gvalue[j] * dweight[j];
  if (DEBUG > 3) 
    fprintf(stdout, "Lvalue %g:%g:%g, Rvalue %g:%g:%g%\n",
            lalpha, lvalue, lfvalue, ralpha, rvalue, rfvalue);

  if        (lvalue * rvalue == 0) {
    if (lvalue == 0) return(lalpha);
    else             return(ralpha);
  } else if (lvalue * rvalue > 0) {
    if (lfvalue > rfvalue) return(lalpha);
    else                   return(ralpha);
  }

  error = fabs(ralpha - lalpha);
  while (error > 1e-5) {

    newalpha=lalpha-lvalue/((rvalue-lvalue)/(ralpha-lalpha));

    /* Evaluate f derivative w.r.t. alpha */
    for (j=0; j<nbweight; j++) newweight[j]=weight[j]+newalpha*dweight[j];
    evaluate(newweight, feature, &newfvalue, gvalue, sztest, nbtpl, nbcluster, 3);
    for (j=0,newvalue=0; j<nbweight; j++) newvalue += gvalue[j] * dweight[j];
    
    if (newvalue == 0) return(newalpha);
    if (newvalue*lvalue<0) {ralpha=newalpha;rvalue=newvalue;rfvalue=newfvalue;}
    else                   {lalpha=newalpha;lvalue=newvalue;lfvalue=newfvalue;}

    if (DEBUG > 3) 
      fprintf(stdout, "Lvalue %g:%g:%g, Rvalue %g:%g:%g%\n",
              lalpha, lvalue, lfvalue, ralpha, rvalue, rfvalue);

    error = fabs(ralpha - lalpha);
  }
 
  if        (lvalue * rvalue == 0) {
    if (lvalue == 0) return(lalpha);
    else             return(ralpha);
  } else if (lvalue * rvalue > 0) {
    if (lfvalue > rfvalue) return(lalpha);
    else                   return(ralpha);
  }

}
