#include "defs.h"

#undef  DEBUG


static void best_first_db_search(
  struct img_set  *img_set,
  struct img_set  *key_set,
  struct node_set *best_ns,
  struct img_info *timg_info);

static void fast_search(
  struct img_info *dimg_info, 
  struct img_info *timg_info, 
  struct node_set *best_ns);

static void full_search(
  struct img_info *dimg_info, 
  struct img_info *timg_info, 
  struct node_set *best_ns,
  int    k);               /* index of image being searched */

float rsize = 1.0;


void fast_db_search(
  struct img_set *img_set,
  struct img_set *key_set,
  struct node_set *best_ns,
  struct img_info *timg_info,
  int    num_to_return)
{
  int k;
  struct img_info *curimg;

  /* Initialize each heap with a full scale search */
  for (k=0; k<img_set->num_img; k++) {

    curimg = &(img_set->img[k]);
#ifdef DEBUG
    fprintf(stdout, "Searching image %s\n", curimg->name); 
    fflush(stdout);
#endif

    /* Initialize node set with full search at coarsest scale */
    full_search(curimg, timg_info, best_ns, k);

  }

}




void list_db_search(
  struct img_set *img_set,
  struct img_set *key_set,
  struct img_info *root,
  struct node_set *best_ns,
  struct img_info *timg_info,
  int num_to_return)
{
  int k;
  struct img_info *curimg;
  struct node p;
  double dist;
  struct node_set buf_ns;


  /* Initialize set of buffer nodes for each image */
  initialize_node_set(&buf_ns);

  /* Initialize each heap with a full scale search */
  curimg = root;
  while (curimg != NULL) {
    k = curimg->ix;

#ifdef DEBUG
    fprintf(stdout, "Searching image %s\n", curimg->name); 
    fflush(stdout);
#endif

    full_search(curimg, timg_info, best_ns, k);

    curimg = curimg->next;
  }

  for (k=0; k<num_to_return && k<img_set->num_img; k++) {
    best_first_db_search(img_set,key_set,best_ns,timg_info);
    if (get_best_from_node_set(best_ns,&p,&dist) == -1) break;
    if (p.key == 1) {fprintf(stdout, "This is a key not image\n"); continue;}
    img_set->img[p.k].order = k;
    insert_into_node_set(&buf_ns,&p,dist);

#ifdef DEBUG
    fprintf(stdout, 
           "Found match %d:%s distance %g\n", k, img_set->img[p.k].name, dist);
#endif 
  }

  /* move buffered nodes to set of best */
  reset_node_set(best_ns);
  for (k=0; k<num_to_return && k<img_set->num_img; k++) {
    if(get_best_from_node_set(&buf_ns,&p,&dist) == -1) break;
    insert_into_node_set(best_ns,&p,dist);
  }

  /* free buffer node set */
  free_node_set(&buf_ns);
}


static void best_first_db_search(
  struct img_set *img_set,
  struct img_set *key_set,
  struct node_set *best_ns,
  struct img_info *timg_info)
{
  int    k;
  struct img_info *curimg, *curkey;
  struct node p;
  double dist;

  /* get best result */
  if (get_best_from_node_set(best_ns, &p, &dist) == -1) return;

  /* process each image in the database */
  while (p.key == 1) {

    /* Initialize each heap with a full scale search */
    curkey = &(key_set->img[p.k]);
    curimg = curkey->child;
    while (curimg != NULL) {
      k = curimg->ix;

      full_search(curimg, timg_info, best_ns, k);

      curimg = curimg->next;
    }
    /* get best result */
    if (get_best_from_node_set(best_ns, &p, &dist) == -1) return;

  }

  /* re-insert best result into heap of best results */
  insert_into_node_set(best_ns,&p,dist);
}


void list_linear_db_search(
  struct img_set *img_set,
  struct img_set *key_set,
  struct img_info *root,
  struct node_set *best_ns,
  struct img_info *timg_info,
  struct img_info **infoqueue)
{
  struct img_info *curimg;
  int    head, tail;

  head = tail = 0;
  curimg = root;
  while (curimg != NULL || head > tail) {

    if (curimg == NULL) {
      curimg = infoqueue[tail++];
    } else if (curimg->type == 1) {
      infoqueue[head++] = curimg->child;
      curimg = curimg->next;
    } else {

      full_search(curimg, timg_info, best_ns, curimg->ix);

      curimg = curimg->next;
    }
  }

}

double direct_search(
  struct img_info *dimg_info, 
  struct img_info *timg_info, 
  struct node_set *ns)
{
  double dist;
  struct node   p;

  reset_node_set(ns);
  full_search(dimg_info, timg_info, ns, dimg_info->ix);

  /* get best results of search */
  get_best_from_node_set(ns,&p,&dist);
  if (dimg_info->type == 1) dist += dimg_info->size * rsize;

  reset_node_set(ns);
  insert_into_node_set(ns, &p, dist);

  return(dist);

}


static void full_search(
  struct img_info *dimg_info, 
  struct img_info *timg_info, 
  struct node_set *best_ns,
  int k)                /* index of image being searched */
{
  struct node p;
  double dist;

  p.k = k;
  p.key = 0;
  dist = feature_distance(dimg_info->feature,timg_info->feature);

  dimg_info->searchcount++;
  if (dimg_info->type == 1) {
    p.key = 1;
    dist -= dimg_info->size * rsize;
  }
  insert_into_node_set(best_ns, &p, dist);

}

void setHeurist(float newrsize)
{

  rsize = newrsize;

}

