#include "defs.h"

#define PI		3.141592653589793


#define	THUMB_SIZE	64
#define SPACING		5

unsigned char myfont[36][5];
unsigned char boundary[8][3]={255, 255, 255, 0, 0, 255, 0, 255, 0, 255, 0, 0,
                              0, 255, 255, 255, 0, 255, 255, 255, 0, 0, 0, 0}; 

void write_fimg(char *fname, struct float_img *img1)
{
  int i, j, l;
  float ***color1;
  float X, Y, Z, R, G, B;
  unsigned char ***color2;
  struct img img2;
  
  img2.row   = img1->row;
  img2.col   = img1->col;
  img2.color = (unsigned char ***) malloc (sizeof (float **) * 3);
  for (l=0; l<3; l++)
    img2.color[l] = (unsigned char **) 
  		get_img(img2.col, img2.row, sizeof(unsigned char));

  color1 = img1->color; color2 = img2.color;
  for (i = 0; i < img2.row; i++)
  for (j = 0; j < img2.col; j++) {
    LAB_to_XYZ(color1[0][i][j], color1[1][i][j], color1[2][i][j], &X, &Y, &Z);
    XYZ_to_RGB(X, Y, Z, &R, &G, &B);
    color2[0][i][j] = R;
    color2[1][i][j] = G;
    color2[2][i][j] = B;
  }

  write_img(fname, &img2, COLOR_JPEG_IMAGE);

  for (l=0; l<3; l++) free_img((void **) img2.color[l]);
  free( (void *) img2.color );

}


void copy_img(struct img *img1, struct img *img2)
{
  int i, j, l;
  unsigned char ***color1,***color2;

  color1 = img1->color; color2 = img2->color;

  for (i = 0; i < img1->row && i < img2->row; i++)
  for (j = 0; j < img1->col && j < img2->col; j++) 
  for (l = 0; l < 3; l++)
    color2[l][i][j] = color1[l][i][j];
}

void prepare_catalog(struct img *all_img, int row, int col, float ratio)
{
  int i, j, l;

  all_img->col = (col * THUMB_SIZE + (col - 1) * SPACING) * ratio; 
  all_img->row = (row * THUMB_SIZE + (row - 1) * SPACING) * ratio;
  all_img->color = (unsigned char ***) malloc (sizeof (float **) * 3);
  for (l=0; l<3; l++)
    all_img->color[l] = (unsigned char **) 
  		get_img(all_img->col, all_img->row, sizeof(unsigned char));

  for (i = 0; i < all_img->row; i++)
  for (j = 0; j < all_img->col; j++) 
  for (l = 0; l < 3; l++)
    all_img->color[l][i][j] = 255;
  
}


void copy_to_catalog(struct img *img1, struct img *img2, int row, int col, float ratio)
{
  int i, j, ii, jj, l, xstart, ystart;
  unsigned char ***color1,***color2;

  color1 = img1->color; color2 = img2->color;
  xstart = (col * (THUMB_SIZE + SPACING) + (THUMB_SIZE - img1->col) / 2)*ratio; 
  ystart = (row * (THUMB_SIZE + SPACING) + (THUMB_SIZE - img1->row) / 2)*ratio;

  for (i = ystart; i < ystart + img1->row*ratio && i < img2->row; i++)
  for (j = xstart; j < xstart + img1->col*ratio && j < img2->col; j++) {
    ii = (i - ystart) / ratio; ii = max(0, min(img1->row-1, ii));
    jj = (j - xstart) / ratio; jj = max(0, min(img1->col-1, jj));
    for (l = 0; l < 3; l++) 
      color2[l][i][j] = color1[l][ii][jj];
  }

  for (l=0; l<3; l++) free_img((void **) img1->color[l]);
  free( (void *)img1->color );

}

void copy_to_catalog2(struct img *img1, struct img *img2, int row, int col, float ratio, int boxcolor)
{
  int i, j, ii, jj, l, xstart, ystart, xstop, ystop;
  unsigned char ***color1,***color2;

  color1 = img1->color; color2 = img2->color;
  xstart = (col * (THUMB_SIZE + SPACING) + (THUMB_SIZE - img1->col) / 2)*ratio; 
  ystart = (row * (THUMB_SIZE + SPACING) + (THUMB_SIZE - img1->row) / 2)*ratio;

  for (i = ystart; i < ystart + img1->row*ratio && i < img2->row; i++)
  for (j = xstart; j < xstart + img1->col*ratio && j < img2->col; j++) {
    ii = (i - ystart) / ratio; ii = max(0, min(img1->row-1, ii));
    jj = (j - xstart) / ratio; jj = max(0, min(img1->col-1, jj));
    for (l = 0; l < 3; l++) 
      color2[l][i][j] = color1[l][ii][jj];
  }

  xstart = (col * (THUMB_SIZE + SPACING))*ratio; 
  ystart = (row * (THUMB_SIZE + SPACING))*ratio;
  xstop  = (col * (THUMB_SIZE + SPACING) + THUMB_SIZE)*ratio; 
  ystop  = (row * (THUMB_SIZE + SPACING) + THUMB_SIZE)*ratio;
  for (i = ystart; i < ystop; i++) {
    for (l = 0; l < 3; l++) color2[l][i][xstart]  = boundary[boxcolor][l];
    for (l = 0; l < 3; l++) color2[l][i][xstop-1] = boundary[boxcolor][l];
  }
  for (j = xstart; j < xstop; j++) {
    for (l = 0; l < 3; l++) color2[l][ystart][j]  = boundary[boxcolor][l];
    for (l = 0; l < 3; l++) color2[l][ystop-1][j] = boundary[boxcolor][l];
  }

  for (l=0; l<3; l++) free_img((void **) img1->color[l]);
  free( (void *)img1->color );

}

int mono_to_color(struct img *data_img, int img_type)
{
  int i, j, l;

  if (img_type == PBM_IMAGE       || img_type == PGM_IMAGE ||
      img_type == GRAY_TIFF_IMAGE || img_type == GRAY_JPEG_IMAGE) {

    data_img->color = (unsigned char ***) malloc (sizeof (float **) * 3);
    for (l=0; l<3; l++)
      data_img->color[l] = (unsigned char **) 
  		get_img(data_img->col, data_img->row, sizeof(unsigned char));
  
    for (i=0; i<data_img->row; i++)
    for (j=0; j<data_img->col; j++)
    for (l=0; l<3; l++)
      data_img->color[l][i][j] = data_img->mono[i][j];

    free_img((void **) data_img->mono);

  } else if (img_type != PPM_IMAGE && img_type != COLOR_TIFF_IMAGE &&
        img_type != COLOR_JPEG_IMAGE) {

    fprintf(stderr,"\nError: can not recognize format of %d\n", img_type); 
    exit(-1);

  }

  return(0);

}

int thumbnail(struct img *data_img, struct img *thumb_img, int maxsize)
{
  int           i, j, ii, jj, l, row, col;
  float         ratio;
  unsigned char ***color1,***color2;

  if (data_img->col > data_img->row) 
    ratio = ((float) data_img->col) / ((float) maxsize);
  else
    ratio = ((float) data_img->row) / ((float) maxsize);
    
  row = (int) (data_img->row / ratio);
  col = (int) (data_img->col / ratio);
  thumb_img->row = (int) (data_img->row / ratio);
  thumb_img->col = (int) (data_img->col / ratio);
  thumb_img->color = (unsigned char ***) malloc (sizeof (unsigned char **) * 3);

  for (l=0; l<3; l++)
    thumb_img->color[l] = (unsigned char **) 
      get_img(thumb_img->col, thumb_img->row, sizeof(unsigned char));
  
  color1 = data_img->color; color2 = thumb_img->color;
  for(i=0; i<thumb_img->row; i++)
  for(j=0; j<thumb_img->col; j++) {
    ii = min((i + 0.5) * ratio, data_img->row-1);
    jj = min((j + 0.5) * ratio, data_img->col-1);
    for(l=0; l<3; l++) 
      color2[l][i][j] = color1[l][ii][jj];
  }

  for (l=0; l<3; l++) free_img((void **) data_img->color[l]);
  free( (void *)data_img->color );

  return(0);

}

int subsampling(struct img *data_img)
{
  int           i, j, ii, jj, l, row, col;
  unsigned char ***color1,***color2;

  row = (int) ((data_img->row + 1)/ 2);
  col = (int) ((data_img->col + 1)/ 2);
  color1 = data_img->color;
  color2 = (unsigned char ***) malloc (sizeof (unsigned char **) * 3);
  for (l=0; l<3; l++)
    color2[l] = (unsigned char **) get_img(col, row, sizeof(unsigned char));
  
  for(i=0; i<row; i++)
  for(j=0; j<col; j++) {
    ii = min(i * 2, data_img->row-1);
    jj = min(j * 2, data_img->col-1);
    for(l=0; l<3; l++) color2[l][i][j] = color1[l][ii][jj];
  }

  for (l=0; l<3; l++) free_img((void **) data_img->color[l]);
  free( (void *)data_img->color );
  data_img->color = color2;

  return(0);

}

void color_space_transform(struct img *img, struct float_img *fimg)
{

  int   i, j, l;
  float R, G, B, X, Y, Z;
  float ratio1, ratio2;
  float Lstar, ustar, vstar, astar, bstar;
  float Yy, Cx, Cz, Hue, Sat; 
  float maxcolor[3], mincolor[3], maxR, maxG, maxB;


  fimg->col = img->col;
  fimg->row = img->row;
  fimg->color = (float ***) malloc (sizeof (float **) * 3);
  for (l=0; l<3; l++)
    fimg->color[l] = (float **) get_img(fimg->col, fimg->row, sizeof(float));

  ratio1 = (float) img->col / (float) fimg->col;
  ratio2 = (float) img->row / (float) fimg->row;
  maxR = maxG = maxB = -1e10;
  for (i=0; i<img->row; i++)
  for (j=0; j<img->col; j++) {
    if (maxR < (float) img->color[0][i][j]) maxR = (float) img->color[0][i][j];
    if (maxG < (float) img->color[1][i][j]) maxG = (float) img->color[1][i][j];
    if (maxB < (float) img->color[2][i][j]) maxB = (float) img->color[2][i][j];
  }
/*
  setwhitepoint(maxR, maxG, maxB);
  fprintf(stdout, "White point %g %g %g\n", maxR, maxG, maxB);
*/
  maxcolor[0] = maxcolor[1] = maxcolor[2] = -1e10;
  mincolor[0] = mincolor[1] = mincolor[2] =  1e10;

  for (i=0; i<fimg->row; i++)
  for (j=0; j<fimg->col; j++) {

    R = degamma((float) img->color[0][(int) (i*ratio2)][(int) (j*ratio1)], (float) GAMMA);
    G = degamma((float) img->color[1][(int) (i*ratio2)][(int) (j*ratio1)], (float) GAMMA);
    B = degamma((float) img->color[2][(int) (i*ratio2)][(int) (j*ratio1)], (float) GAMMA);

    switch (COLORSPACE) {
    case RGB:
      fimg->color[0][i][j] = R;
      fimg->color[1][i][j] = G;
      fimg->color[2][i][j] = B;
      break;

    case XYZ:
      RGB_to_XYZ(R, G, B, &X, &Y, &Z);

      fimg->color[0][i][j] = X;
      fimg->color[1][i][j] = Y;
      fimg->color[2][i][j] = Z;
      break;

    case LUV:
      RGB_to_XYZ(R, G, B, &X, &Y, &Z);
      XYZ_to_LUV(X, Y, Z, &Lstar, &ustar, &vstar);

      fimg->color[0][i][j] = Lstar;
      fimg->color[1][i][j] = ustar;
      fimg->color[2][i][j] = vstar;
      break;

    case LAB:
      RGB_to_XYZ(R, G, B, &X, &Y, &Z);
      XYZ_to_LAB(X, Y, Z, &Lstar, &astar, &bstar);

      fimg->color[0][i][j] = Lstar;
      fimg->color[1][i][j] = astar;
      fimg->color[2][i][j] = bstar;
      break;

    case SLAB:
      RGB_to_XYZ(R, G, B, &X, &Y, &Z);
      XYZ_to_SLAB(X, Y, Z, &Yy, &Cx, &Cz);

      fimg->color[0][i][j] = Yy;
      fimg->color[1][i][j] = Cx;
      fimg->color[2][i][j] = Cz;
      break;

    case LHS:
      RGB_to_XYZ(R, G, B, &X, &Y, &Z);
      XYZ_to_LHS(X, Y, Z, &Lstar, &Hue, &Sat);

      fimg->color[0][i][j] = Lstar;
      fimg->color[1][i][j] = Hue;
      fimg->color[2][i][j] = Sat;
      break;

    default:
      break;
    }
    if (fimg->color[0][i][j] > maxcolor[0]) maxcolor[0] = fimg->color[0][i][j];
    if (fimg->color[0][i][j] < mincolor[0]) mincolor[0] = fimg->color[0][i][j];
    if (fimg->color[1][i][j] > maxcolor[1]) maxcolor[1] = fimg->color[1][i][j];
    if (fimg->color[1][i][j] < mincolor[1]) mincolor[1] = fimg->color[1][i][j];
    if (fimg->color[2][i][j] > maxcolor[2]) maxcolor[2] = fimg->color[2][i][j];
    if (fimg->color[2][i][j] < mincolor[2]) mincolor[2] = fimg->color[2][i][j];
  }
/*
  fprintf(stdout, "Color Range : <%5.1f:%5.1f><%5.1f:%5.1f><%5.1f:%5.1f>\n",
         maxcolor[0], mincolor[0], maxcolor[1], mincolor[1], 
         maxcolor[2], mincolor[2]);

  for (l=0; l<3; l++) free_img((void **) img->color[l]);
  free( (void *)img->color );
*/

}

double image_distance(struct float_img *img1, struct float_img *img2)
{
  int i, j, l;
  double L;
  float ***color1, ***color2;

  L = 0;
  color1 = img1->color; color2 = img2->color;

  for (i = 0; i < img1->row && i < img2->row; i++)
  for (j = 0; j < img1->col && j < img2->col; j++) 
  for (l = 0; l < 3; l++)
    L += fabs(color2[l][i][j] - color1[l][i][j]);
  L /= (img1->row * img1->col);

  return(L);
}


