#!/usr/local/bin/wish
#Global variable
# map: pointer from icon position to image index
# selection: selection flag for add/remove from the list at right hand side
# pratio:  prune percentage



proc makedir {pathname} {

    if {[file isdirectory $pathname]} {
        return $pathname
    } elseif {[file exists $pathname]} {
        puts "non-directory $pathname already exists"
    } else {
        # Recurse to create intermediate directories
        makedir [file dirname $pathname]
        file mkdir $pathname
        return $pathname
    }


}


#Prune: prune the database into a subset, and build a pyramid base on default
#       weighting
proc prunelist { } {
    global appFp subset pratio vpoption level istart jstart weight selection map 

    set relevance  ""
    set rcount     0
    for {set i 0} {$i < 32} {incr i 1} {
      for {set j 0} {$j < 2} {incr j 1} {
        set pos [format "1x%dx%d" $i $j]
        if [info exist map($pos)] {
          set relevance [format "%s %s" $relevance $map($pos)]
          set rcount [expr $rcount + 1]
        }
      }
    }

    if {$rcount > 0} {
      set allweight [format "%s" $weight(0) ]
      for {set i 1} {$i < 9} {incr i 1} {
        set allweight [format "%s:%s" $allweight $weight($i) ]
      }

      set kth [expr $pratio*($rcount+1)+0.5]
      puts $appFp "prunelist $kth $relevance"
      flush $appFp
      gets $appFp line

      set subset 1
      puts $appFp "buildpym 0 $subset $relevance"
      flush $appFp
      gets $appFp line

      set level 1
      set istart -2
      set jstart -2
      browse

    }
 
}


#Rebuild Pyramid according to optimized color/feature weighting
proc rebuild {pymtype} {
    global appFp subset level istart jstart weight map 

    set relevance  ""
    set rcount     0
    for {set i 0} {$i < 32} {incr i 1} {
      for {set j 0} {$j < 2} {incr j 1} {
        set pos [format "1x%dx%d" $i $j]
        if [info exist map($pos)] {
          set relevance [format "%s %s" $relevance $map($pos)]
          set rcount [expr $rcount + 1]
        }
      }
    }

    set allweight [format "%s" $weight(0) ]
    for {set i 1} {$i < 9} {incr i 1} {
      set allweight [format "%s:%s" $allweight $weight($i) ]
    }

    puts $appFp "buildpym $pymtype $subset $relevance"
    flush $appFp
    gets $appFp line

    set level 1
    set istart -2
    set jstart -2
    browse
 

}

#Full Browsing: set current database to full database
proc full { } {
    global subset level istart jstart 

    if {$subset == 0} {
      return
    }

    set subset 0
    set level  2
    set istart 0
    set jstart 0
    browse

 
}

proc add { } {
    global appFp subset map selection pratio

    for {set i 0} {$i < 8} {incr i 1} {
      for {set j 0} {$j < 8} {incr j 1} {
        set pos [format "0x%dx%d" $i $j]
        if [info exist map($pos)] {
          if {$selection($pos) == 1} {
            .icon($pos) configure -borderwidth 0
            set selection($pos) 0
            set notfound 1
            for {set m 0} {$m < 32 && $notfound == 1} {incr m 1} {
              for {set n 0} {$n < 2 && $notfound == 1} {incr n 1} {
                set Rpos [format "1x%dx%d" $m $n]
                if [info exist map($Rpos)] {
                  if {$map($Rpos) == $map($pos)} {
                    set notfound 0
                  }
                }
              }
            }
            for {set m 0} {$m < 32 && $notfound == 1} {incr m 1} {
              for {set n 0} {$n < 2 && $notfound == 1} {incr n 1} {
                set Rpos [format "1x%dx%d" $m $n]
                if ![info exist map($Rpos)] {
                  set map($Rpos) $map($pos)
                  show $Rpos 
                  set notfound 0
                }
              }
            }
          }
        }
      }
    }

    set nbrelevance 0
    for {set m 0} {$m < 32} {incr m 1} {
      for {set n 0} {$n < 2} {incr n 1} {
        set Rpos [format "1x%dx%d" $m $n]
        if [info exist map($Rpos)] {
          set nbrelevance [expr $nbrelevance + 1]
        }
      }
    }
    .pratio configure -from [expr 1.0 / ($nbrelevance + 1.0)] -to [expr $nbrelevance / ($nbrelevance + 1.0)] -resolution [expr 1.0 / ($nbrelevance + 1.0)]
    set pratio [expr $nbrelevance * 0.5 / ($nbrelevance + 1.0)]

}

proc remove { } {
    global map selection pratio

    for {set i 0} {$i < 32} {incr i 1} {
      for {set j 0} {$j < 2} {incr j 1} {
        set pos [format "1x%dx%d" $i $j]
        if [info exist map($pos)] {
          if {$selection($pos) == 1} {
            .icon($pos) configure -borderwidth 0
            set selection($pos) 0
            unset map($pos)
            image($pos) blank
          }
        }
      }
    }

    set nbrelevance 0
    for {set m 0} {$m < 32} {incr m 1} {
      for {set n 0} {$n < 2} {incr n 1} {
        set Rpos [format "1x%dx%d" $m $n]
        if [info exist map($Rpos)] {
          set nbrelevance [expr $nbrelevance + 1]
        }
      }
    }
    .pratio configure -from [expr 1.0 / ($nbrelevance + 1.0)] -to [expr $nbrelevance / ($nbrelevance + 1.0)] -resolution [expr 1.0 / ($nbrelevance + 1.0)]
    set pratio [expr $nbrelevance * 0.5 / ($nbrelevance + 1.0)]

}

#right button clicked: mark image as selected
proc select {r i j} {
    global appFp map selection 

    set pos [format "%dx%dx%d" $r $i $j]
    if [info exist map($pos)] {

      puts $appFp "filename $map($pos)"
      flush $appFp
      gets $appFp line
      .mesg configure -text [lindex $line 2]

      if {$selection($pos) == 0} {
        .icon($pos) configure -borderwidth 2
        set selection($pos) 1
      } else {
        .icon($pos) configure -borderwidth 0
        set selection($pos) 0
      }
    }

}

#right button clicked: mark image as selected
proc locate {r i j} {
    global appFp map subset level istart jstart

    set pos [format "%dx%dx%d" $r $i $j]
    if [info exist map($pos)] {

      puts $appFp "filename $map($pos)"
      flush $appFp
      gets $appFp line
      .mesg configure -text [lindex $line 2]

      puts $appFp "locate $subset $map($pos)"
      flush $appFp
      gets $appFp line

      scan [lindex $line 2] {%d} level
      scan [lindex $line 3] {%d} i
      scan [lindex $line 4] {%d} j
       
      set istart [expr $i - 3]
      set jstart [expr $j - 3]
      browse
 
    }

}

#Left button click: search/browse according program mode
proc clicking {i j} {
    global pmode map

    set pos [format "0x%dx%d" $i $j]
    if [info exist map($pos)] {
      if {$pmode == "browse"} {
        zoomin $i $j
      } else {
	search $i $j
      }
    }
}

# image i, j is clicked under search mode
proc search {i j} {
    global appFp subset map weight

    set allweight [format "%s" $weight(0) ]
    for {set k 1} {$k < 9} {incr k 1} {
      set allweight [format "%s:%s" $allweight $weight($k) ]
    }

    set pos [format "0x%dx%d" $i $j]
    puts $appFp "match $subset $map($pos)"
    flush $appFp
 
    gets $appFp line
    display $line 2
 

}

#image i, j is clicked under browser mode
proc zoomin {i j} {
    global map subset level istart jstart

    set pos [format "0x%dx%d" $i $j]

    if [info exist map($pos)] {
        incr level 1
        set istart [expr ($i + $istart) * 2 - 3]
        set jstart [expr ($j + $jstart) * 2 - 3]
        browse
    }

}

#zoom out button is clicked
proc zoomout {} {
    global level istart jstart 

    if {$level <= 1} {
        set level 1
        set istart -2
        set jstart -2
    } else {
        incr level -1
        set istart [expr [expr int ($istart + 1) / 2] - 2]
        set jstart [expr [expr int ($jstart + 1) / 2] - 2]
    }
    browse

}

#pan button is clicked
proc pan {direction} {
    global level istart jstart

    set size [expr pow(2, [expr $level + 1])]
    switch -exact -- $direction {
        up    {if {$istart >= -4}        {incr istart -4}}
        left  {if {$jstart >= -4}        {incr jstart -4}}
        right {if {$jstart <= $size - 4} {incr jstart 4}}
        down  {if {$istart <= $size - 4} {incr istart 4}}
    }
    browse

}

#browse a view window of pyramid
proc browse {} {
    global appFp subset level istart jstart

    puts $appFp "browse $subset $level $istart $jstart"
    flush $appFp
 
    gets $appFp line
    display $line 2
 
}

#display the search result
proc display {line offset} {
    global appFp subset selection map image 

    puts $line
    for {set i 0} {$i < 8} {incr i 1} {
      for {set j 0} {$j < 8} {incr j 1} {
        set pos [format "0x%dx%d" $i $j]
        set selection($pos) 0
        .icon($pos) configure -borderwidth 0

        set k [expr $i * 8 + $j + $offset]
        scan [lindex $line $k] {%d} index
        if {$index == -1} {
          if [info exist map($pos)] {
            unset map($pos) 
            image($pos) blank
          }
        } else {
          set map($pos) $index
          show $pos 
        }
      }
    }
}

#show image
proc show {pos} {
    global appFp image map

    if ![info exist map($pos)] {
      image($pos) blank

    } else {
	    image create photo tmpimage 
	    puts $appFp "extract $map($pos) 64"
	    flush $appFp

	    gets $appFp line
	    tmpimage read "extract.ppm"
	    image($pos) blank
	    image($pos) copy tmpimage -zoom 1 -to [expr (64 - [image width tmpimage])/2] [expr (64 - [image height tmpimage])/2] 
	    image delete tmpimage 
	}

}


#Left button click: search/browse according program mode
proc setpmode {mode} {
    global pmode level istart jstart

    if {$pmode == $mode} {
      return;
    } else {
      if {$mode == "browse"} {
        set pmode browse
        browse
      } else {
        set pmode search
      }
    }
}
#Load Class
proc loadclass {index} {
    global classfile classentry classname classmap nbclass curclass name map kframe selection pratio

    .mesg configure -text [format "%d %d" $index $nbclass]

    if {$index < $nbclass} {

        set curclass $index
        set classentry $classname($curclass)
        .classentry configure -state disabled

        set k 0
        for {set i 0} {$i < 32} {incr i 1} {
          for {set j 0} {$j < 2} {incr j 1} {
            set pos [format "1x%dx%d" $i $j]
            set ix  [format "%dX%d" $curclass $k]
            .icon($pos) configure -borderwidth 0
            set selection($pos) 0

            if [info exist map($pos)] {
                unset map($pos)
                image($pos) blank
            }
            if {$classmap($ix) != "-1"} {
                set map($pos) $classmap($ix)
                show $pos
            }
            incr k 1
          }
        }
       
        set nbrelevance 0
        for {set m 0} {$m < 32} {incr m 1} {
          for {set n 0} {$n < 2} {incr n 1} {
            set pos [format "1x%dx%d" $m $n]
            if [info exist map($pos)] {
              set nbrelevance [expr $nbrelevance + 1]
            }
          }
        }
        .pratio configure -from [expr 1.0 / ($nbrelevance + 1.0)] -to [expr $nbrelevance / ($nbrelevance + 1.0)] -resolution [expr 1.0 / ($nbrelevance + 1.0)]
        set pratio [expr $nbrelevance * 0.5 / ($nbrelevance + 1.0) ]

    }


}

#Save Class
proc saveclass {} {
    global classfile classentry classname classmap nbclass curclass map 

    if {$nbclass == 0} {return}

    set classname($curclass) $classentry
    .classentry configure -state disabled

    for {set i 0} {$i < 64} {incr i 1} {
          set ix  [format "%dX%d" $curclass $i]
          set classmap($ix) "-1"
    }

    set k 0
    for {set i 0} {$i < 32} {incr i 1} {
      for {set j 0} {$j < 2} {incr j 1} {
          set pos [format "1x%dx%d" $i $j]
          set ix  [format "%dX%d" $curclass $k]

          if [info exist map($pos)] {
              set classmap($ix) $map($pos)
              incr k 1
          }
      }
    }

    .classmenu.classlist delete 2 [expr $nbclass+1]

    set class   [open $classfile w]
    puts $class $nbclass
    for {set i 0} {$i < $nbclass} {incr i 1} {
      .classmenu.classlist add command -label $classname($i) -command "loadclass $i" 
      puts $class $classname($i)
      set clist ""
      for {set j 0} {$j < 64} {incr j 1} {
        set ix  [format "%dX%d" $i $j]
        set clist [format "%s %s" $clist $classmap($ix)]
      }
      puts $class $clist
    }
    close $class

        
}

#New Class
proc newclass {} {
    global classentry classname classmap nbclass curclass

    set curclass $nbclass
    set classentry "NoName"

    set classname($curclass) $classentry
    for {set i 0} {$i < 64} {incr i 1} {
          set ix  [format "%dX%d" $curclass $i]
          set classmap($ix) -1
    }

    .classmenu.classlist add command -label $classname($nbclass) -command "loadclass $nbclass"

    incr nbclass 1
    .classentry configure -state normal

        
}

proc addlist {} {
    global appFp subset

    if {$subset == 1} {
      set subset 0
      set level 1
      set istart -2
      set jstart -2
    }

    uplevel [list source getfile.tcl]
    tkwait window .get
    puts $appFp "addlist list_of_images"
    flush $appFp
 
    gets $appFp line
    browse
 

}

proc quit { } {
    global appFp

    puts $appFp "exit"
    flush $appFp
    close $appFp
    exit
 
}



wm geometry . 800x544+0+0

set workdir "db_folderSimple"
makedir $workdir
makedir "$workdir/feature"
makedir "$workdir/SFLEXIBLE"

set cmd "|active.exe $workdir"
set appFp [open  $cmd "r+"]
wm protocol . WM_DELETE_WINDOW quit

#Initialize icon images
for {set i 0} {$i < 8} {incr i 1} {
  for {set j 0} {$j < 8} {incr j 1} {
    set pos [format "0x%dx%d" $i $j]
    image create photo image($pos) -width 64 -height 64
    button .icon($pos) -image image($pos) -command "clicking $i $j"
    bind   .icon($pos) <Button-2> "locate 0 $i $j"
    bind   .icon($pos) <Button-3> "select 0 $i $j"
    place  .icon($pos) -x [expr $j * 68]  -y [expr $i * 68]
    .icon($pos) configure -borderwidth 0
  }
}

#program mode label/selection definition

menubutton .pmodemenu -text "MODE" -menu .pmodemenu.pmodelist -relief raised
place .pmodemenu -x 544 -y 10 -width 50
menu .pmodemenu.pmodelist -tearoff 0
.pmodemenu.pmodelist add command -label Search -command "setpmode search" 
.pmodemenu.pmodelist add command -label Browse -command "setpmode browse" 
entry  .pmodeentry -textvar pmode -width 10 -justify left -state disabled
place  .pmodeentry -x 594 -y 10
set pmode browse

#Pseudo Semantic menu
menubutton .classmenu -text "Semantic Class" -menu .classmenu.classlist -relief raised
place .classmenu -x 544 -y 40 -width 100
menu .classmenu.classlist -tearoff 0
.classmenu.classlist add command -label New -command "newclass" 
.classmenu.classlist add separator

set classfile [format "%s/class" $workdir]
if {![file exists $classfile]} {
    set class   [open $classfile w]
    puts $class 0
    close $class
}

set class   [open $classfile]
gets $class line
scan $line {%d} nbclass
for {set i 0} {$i < $nbclass} {incr i 1} {
  gets $class line
  set classname($i) $line
  .classmenu.classlist add command -label $classname($i) -command "loadclass $i"
  gets $class line
  for {set j 0} {$j < 64} {incr j 1} {
    set ix  [format "%dX%d" $i $j]
    set classmap($ix)  [lindex $line $j]
  }
}
close $class

entry  .classentry -textvar classentry -width 10 -justify left -state disabled
place  .classentry -x  594 -y 70
set classentry ""
set curclass $nbclass

button .saveclass -text "Save"  -command "saveclass" -justify center
place  .saveclass -x  544 -y 70 -width 50

#list handle button definition
button .add -text "Add" -command add -justify center
place  .add -x 544 -y 110 -width 50

button .remove -text "Remove" -command remove -justify center
place  .remove -x 594 -y 110 -width 50

for {set i 0} {$i < 9} {incr i 1} {
  set weight($i) 1
}

#Movement Buttion Definition, to be changed into arrow icon
image create photo arrowup    -file [file join images arrowup.gif]
image create photo arrowdown  -file [file join images arrowdown.gif]
image create photo arrowleft  -file [file join images arrowleft.gif]
image create photo arrowright -file [file join images arrowright.gif]
image create photo arrowtop   -file [file join images arrowtop.gif]
button .panup    -image arrowup -command "pan up"    -justify center 
place  .panup    -x  578 -y 147 -width 33
button .panleft  -image arrowleft  -command "pan left"  -justify center
place  .panleft  -x  544 -y 180 -width 33
button .zoomout  -image arrowtop   -command "zoomout"   -justify center
place  .zoomout  -x  578 -y 180 -width 33
button .panright -image arrowright -command "pan right" -justify center
place  .panright -x  612 -y 180 -width 33
button .pandown  -image arrowdown  -command "pan down"  -justify center
place  .pandown  -x  578 -y 213 -width 33

#Function Button Definition
scale .pratio -from 0 -to 1 -resolution 1 -length 95 -variable pratio -orient horizontal -label "Accuracy: m %"
place  .pratio -x 544 -y 250 
set    pratio 1

button .addlist -text "Add Images" -command addlist -justify center
place  .addlist -x 544 -y 340 -width 100

button .prune -text "Prune" -command prunelist -justify center
place  .prune -x 544  -y 370 -width 50

button .rebuild -text "Rebuild" -command "rebuild 0" -justify center
place  .rebuild -x 594 -y 370 -width 50

button .optimize -text "Optimize" -command "rebuild 1" -justify center
place  .optimize -x 544 -y 400 -width 50

button .wormhole -text "Wormhole" -command "rebuild 2" -justify center
place  .wormhole -x 594 -y 400 -width 50

button .full -text "Full" -command full -justify center
place  .full -x 544 -y 430 -width 50

button .exit -text EXIT -command quit -justify center
place  .exit -x 594 -y 430 -width 50

# message box
gets $appFp line
label  .info -wraplength 90 -justify center -text "DB: [lindex $line 1] images"
place  .info -x 544 -y 470 

label  .mesg -wraplength 90 -justify center -text "L: Browse/Search\nR: Select/Query"
place  .mesg -x 544 -y 490 

frame  .f
place  .f -x 644 -y 0
canvas .f.r -width 136 -height 544 -yscrollcommand [list .f.yscroll set]
scrollbar .f.yscroll -orient vertical -command [list .f.r yview]
pack .f.yscroll -side right -fill y
pack .f.r -side left -fill both -expand true

set f [frame .f.r.f]
.f.r create window 0 0 -anchor nw -window $f
#Initialize relevance images
frame .f.r.f.f -width 136 -height 2176
for {set i 0} {$i < 32} {incr i 1} {
  for {set j 0} {$j < 2} {incr j 1} {

    set pos [format "1x%dx%d" $i $j]
    set selection($pos) 0

    image create photo image($pos) -width 64 -height 64
    button .icon($pos) -image image($pos)
    bind   .icon($pos) <Button-2> "locate 1 $i $j"
    bind   .icon($pos) <Button-3> "select 1 $i $j"
    place  .icon($pos) -in .f.r.f.f -x [expr $j * 68]  -y [expr $i * 68]
    .icon($pos) configure -borderwidth 0

  }
}
pack .f.r.f.f -in .f.r.f -side top
.f.r config -scrollregion "0 0 136 2176"
.f.r config -width 136 -height 544

set subset 0
set level 1
set istart -2
set jstart -2
browse

