#!/usr/bin/wish
# browse --
# This script generates a directory browser, which lists the working
# directory and allows you to open files or subdirectories by
# double-clicking.

# The procedure below is invoked to open a browser on a given file;  if the
# file is a directory then another instance of this program is invoked; if
# the file is a regular file then the Mx editor is invoked to display
# the file.

proc getdir {pathname} {

#UNIX
#    return [split [exec ls -a $pathname] '\n']

#WINDOW NT
#    regsub -all / $pathname \\ path
#    return [split [exec cmd.exe /c dir /b /o:gne $path] '\n']

#WINDOW 9X
#    regsub -all / $pathname \\ path
#    return [split [exec command.com /c dir /b /o:gne $path] '\n']
     set wd [pwd]
     cd $pathname
     set directory [lsort -ascii [glob -nocomplain *]]
     cd $wd
     return $directory

}


proc ScrolledListbox {parent args} {
    frame $parent

    #create listbox attached to scrollbars, pass thru $arg
    eval {listbox $parent.list \
          -selectmode extended \
          -yscrollcommand [list $parent.sy set] \
          -xscrollcommand [list $parent.sx set]} $args

    #create scrollbars attached to the listbox
    scrollbar $parent.sx -orient horizontal -command [list $parent.list xview]
    scrollbar $parent.sy -orient vertical   -command [list $parent.list yview]

    #Arrange them in the parent frame
    pack $parent.sx -side bottom -fill x
    pack $parent.sy -side right  -fill y

    #Pack to allow for resizing
    pack $parent.list -side left -fill both -expand true

    return $parent.list
}

proc subadd {pathname} {
    global getFp

    if {[file isdirectory $pathname]} {
        # Recurse to create intermediate directories
        foreach i [getdir $pathname] { 
            if {[string compare $i ""] == 0} {continue;}
            if {[string compare $i "."] == 0} {continue;}
            if {[string compare $i ".."] == 0} {continue;}
            set file $pathname/$i
            if ![file exist $file] { continue; }
            subadd $file
        }
        return $pathname
    } elseif {[file exists $pathname]} {
        if {[file size $pathname] == 0} { continue; }
        if ![file exist $pathname] { continue; }
        puts $getFp $pathname
        flush $getFp
        gets $getFp line
        set found [lindex $line 1]
        if {$found == "OK"} {
           .get.select.list insert end $pathname
        }
        return $pathname
    } else {
        return $pathname
    }



}

proc addselection {} {
    global env path

    foreach i [.get.list.list curselection] { 
       set file $path/[.get.list.list get $i]
       subadd $file
    }
}

proc removeselection {} {
    global env path

    foreach i [.get.select.list curselection] { 
       .get.select.list delete $i
    }
}

proc createlist {} {
    global getFp

    set listfile [open list_of_images w]
    puts $listfile [.get.select.list index end]
    foreach i [.get.select.list get 0 end] {
        puts $listfile $i
    }
    close $listfile

    puts $getFp "exit"
    flush $getFp
    close $getFp
    destroy .get
 
}

proc browsedir {dir file} {
    global getFp
    global env path

    if {[string compare $file "."] == 0} {return}
    if {[string compare $file ".."] == 0} {
        set file [file dirname $dir]
    } else {
        set file [file join $dir $file]
        if {[file type $file] == "link"} {
            set file [file readlink $file]
        }
    }
    if [file isdirectory $file] {
        .get.list.list delete 0 end
        .get.list.list insert end "."
        .get.list.list insert end ".."

        foreach i [getdir $file] { 
            if {[string compare $i ""] == 0} {continue; }
            if {[string compare $i "."] == 0} {continue; }
            if {[string compare $i ".."] == 0} {continue; }
            if ![file exist [file join $file $i]] { continue; }
            if [file isdirectory [file join $file $i]]  { 
               .get.list.list insert end $i 
           }
        }
        foreach i [getdir $file] { 
            if {[string compare $i ""] == 0} {continue; }
            if {[string compare $i "."] == 0} {continue; }
            if {[string compare $i ".."] == 0} {continue; }
            if ![file exist [file join $file $i]] { continue; }
            if ![file isdirectory [file join $file $i]]  { 
               if {[file size [file join $file $i]] == 0} { continue; }
               puts $getFp [file join $file $i]
               flush $getFp
               gets $getFp line
               set found [lindex $line 1]
               if {$found == "OK"} { .get.list.list insert end $i }
           }
        }
        foreach i [file volume] {
            .get.list.list insert end $i
        }
        set path $file
        .get.path configure -text $path
        # clear list and then add list
        # Fill the listbox with a list of all the files in the directory (run
        # the "ls" command to get that information).
    } else {
        #add file to database
        if {[file size $file] == 0} { continue; }
        puts $getFp $file
        flush $getFp
        gets $getFp line
        set found [lindex $line 1]
        if {$found == "OK"} { 
           .get.select.list insert end $file
        }
    }
}


proc quitget { } {
    global getFp

    set listfile [open list_of_images w]
    puts $listfile 0
    close $listfile

    puts $getFp "exit"
    flush $getFp
    close $getFp
    destroy .get
 
}

toplevel .get
wm geometry .get 500x450+0+0

set cmd "|isimage.exe"
set getFp [open  $cmd "r+"]

set path [pwd]
set file {.}

message .get.path -text $path -aspect 2000
place .get.path -x 0 -y 0 -width 500

message .get.help -text "Add: adds selected files or folders to image database\nRemove: removes selected file or folder from the list of images to be added\nAll images are added to database at termination of program" -aspect 2000
place .get.help -x 0 -y 20 -width 500

button .get.add -text "Add" -command addselection -justify center
place  .get.add -x 200 -y 150 -width 100

button .get.remove -text "Remove" -command removeselection -justify center
place  .get.remove -x 200 -y 250 -width 100

button .get.ok     -text ok     -command createlist
place  .get.ok -x 100 -y 400 -width 100

button .get.cancel -text cancel -command quitget
place .get.cancel -x 300 -y 400 -width 100

#
# Create a scrollbar on the right side of the main window and a listbox
# on the left side.

ScrolledListbox .get.list  -width 30 -height 20 
place .get.list -x 0 -y 100 -width 200 -height 300

ScrolledListbox .get.select -width 30 -height 20 
place .get.select -x 300 -y 100 -width 200 -height 300

.get.list.list insert end "."
.get.list.list insert end ".."

foreach i [getdir $path] {
    if {[string compare $i ""] == 0} {continue; }
    if {[string compare $i "."] == 0} {continue; }
    if {[string compare $i ".."] == 0} {continue; }
    if ![file exist $i] { continue; }
    if [file isdirectory $i] { 
      .get.list.list insert end $i 
    }
}

foreach i [getdir $path] {
    if {[string compare $i ""] == 0} {continue; }
    if {[string compare $i "."] == 0} {continue; }
    if {[string compare $i ".."] == 0} {continue; }
    if ![file exist $i] { continue; }
    if ![file isdirectory $i] { 
      if {[file size $i] == 0} { continue; }
      puts $getFp [file join $path $i]
      flush $getFp
      gets $getFp line
      puts $line
      set found [lindex $line 1]
      if {$found == "OK"} { .get.list.list insert end $i }
    }
}

foreach i [file volume] {
      .get.list.list insert end $i
}

# Set up bindings for the browser.

bind .get.list.list <Double-Button-1> {foreach i [.get.list.list curselection] {browsedir $path [.get.list.list get $i]}}
bind .get.select.list <Double-Button-1> {foreach i [.get.select.list curselection] {.select.list delete $i}}

